# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Module that handles the update of call sources."""


import argparse
import re
import sys
from typing import List, Optional
from ruamel.yaml import YAML
from pathlib import Path


PARSER_HELP = "Update call sources for method calls in source files."


def configure_parser(parser: argparse.ArgumentParser):
    """Configure the parser for the update call sources command."""
    parser.add_argument("interface", help="Interface specification dict", type=Path)
    parser.add_argument("src_dir", help="Path to source file directory", type=Path)
    parser.add_argument(
        "-p",
        "--project-config",
        type=Path,
        default=None,
        help="Path to project config json file",
    )
    parser.set_defaults(func=update_call_sources_cli)


def update_call_sources_cli(args: argparse.Namespace):
    """CLI function for updating call sources."""
    method_config = read_project_config(args.project_config)
    with open(args.interface, encoding="utf-8") as interface_file:
        yaml = YAML(typ='safe', pure=True)
        adapter_spec = yaml.load(interface_file)
    update_call_sources(args.src_dir, adapter_spec, method_config)


def main(argv: Optional[List[str]] = None):
    """ Main function for stand alone execution."""
    parser = argparse.ArgumentParser(description=PARSER_HELP)
    configure_parser(parser)
    args = parser.parse_args(argv)
    args.func(args)


def read_project_config(project_config_path):
    """ Reads project config file and extract method specific settings if they are present.

    Args:
        project_config_path (Path): path to the ProjectCfg.json file
    Returns:
        method_config (dict): dictionary of method related configs.
    """
    project_info = {}
    if project_config_path is not None:
        with project_config_path.open() as config_file:
            yaml = YAML(typ='safe', pure=True)
            config = yaml.load(config_file)
            project_info = config["ProjectInfo"]
    method_config = {
        "adapter_declarations_change": project_info.get("adapterDeclarations", None)
    }
    method_config["method_call_wrapper"] = {
        "pre": project_info.get("methodPreCallWrapper", ""),
        "post": project_info.get("methodPostCallWrapper", "")
    }
    for key, value in method_config["method_call_wrapper"].items():
        if value != "":
            method_config["method_call_wrapper"][key] = value + "\n"

    return method_config


def is_method_adapter(adapter):
    """ Check if adapter has methods in it

    Args:
        adapter (dict): adapter specification
    Returns:
        methods_in_adapter (bool): true if adapter contains methods,
        false otherwise
    """
    methods_in_adapter = "methods" in adapter and len(adapter["methods"]) > 0
    return methods_in_adapter


def update_call_sources(src_dir, adapter_spec, method_config):
    """ Update the source files for specified method calls with
    adapter function calls that trigger the methods.

    Args:
        src_dir (Path): path to folder for method call sources
        adapter_spec (list): adapter specifications with methods
        method_config (dict): project specific method configs
    """

    method_adapters = [a for a in adapter_spec if is_method_adapter(a)]
    for adapter in method_adapters:
        for method in adapter["methods"]:
            method_src = src_dir / (method["name"] + ".c")
            with method_src.open("r+") as src_file:
                old_src = src_file.read()
            new_src = generate_src_code(adapter, method, old_src, method_config)
            method_src.unlink()
            with open(method_src.with_suffix(".cpp"), "w", encoding="utf-8") as dst_file:
                dst_file.write(new_src)
            method_header = src_dir / (method["name"] + ".h")
            with method_header.open("r+") as header_file:
                old_header = header_file.read()
                new_header = generate_header_code(method_config, old_header)
                header_file.seek(0)
                header_file.write(new_header)
                header_file.truncate()


def generate_header_code(method_config, old_header):
    """ Change header code to include project specific adapter wrapper.

    Args:
        method_config (dict): project specific method settings
        old_header (string): header source code
    Returns:
        new_header (string): modified header source code

    """
    adapter_pattern = r'(?<=#include ")adapter_wrapper.hh'
    if method_config["adapter_declarations_change"]:
        new_header = re.sub(
            adapter_pattern, method_config["adapter_declarations_change"], old_header
        )
        return new_header
    return old_header


def generate_src_code(adapter, method, old_src, method_config):
    """ Generate the method call source code to trigger the method call

    Args:
        adapter (dict): adapter specification
        method (dict): method specification
        old_src (string): method call source code with mock call
        method_config (dict): project specific method settings
    Returns:
        new_src (string): method call source code with adapter function call
        that triggers method

    """
    call_pattern = (
        r"    \/\* CSP method call\n"
        r"(.*)ADAPTER->METHOD(.*)"
        r"    CSP method call end\*\/\n"
    )
    dummy_pattern = (
        r"    \/\* C dummy call\*\/" r".*" r"    \/\* C dummy call end\*\/\n"
    )
    comment = (
        r"\/\* Used for running spm without csp, such as silver\n"
        r"   This code should be replaced when using csp. \*\/\n"
    )
    name = method["name"]
    adapter_name = [adapter["name"].title()]
    namespace = []
    if method["namespace"] != "":
        namespace = [method["namespace"]]
    adapter = "::".join(namespace + adapter_name) + "->" + name
    pre_call = method_config["method_call_wrapper"]["pre"]
    post_call = method_config["method_call_wrapper"]["post"]
    sans_comment = re.sub(comment, "", old_src)
    sans_dummy_call = re.sub(dummy_pattern, "", sans_comment, flags=re.DOTALL)
    new_src = re.sub(
        call_pattern, pre_call + r"\1" + adapter + r"\2" + post_call, sans_dummy_call, flags=re.DOTALL
    )
    return new_src


if __name__ == "__main__":
    main(sys.argv[1:])



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Module containing classes for the signal interfaces report."""
from collections import defaultdict
from powertrain_build.html_report import HtmlReport


class SigIfHtmlReportAll(HtmlReport):
    """Signal interface html report.

    Inherits class :doc:`HtmlReport <html_report>`.
    """

    __intro = """
    <h2 id="introduction">Introduction</h2>
    <p>This documents lists the signal interfaces.</p>
    """

    __signal_info = """<h2 class="nocount">Table of contents</h2>
    <ol>
    <li><a href="#signal_details">Detailed signal information</a></li>
    <li><a href="#signal_index">Signal index</a></li>
    </ol>
    """

    __table_unused = """
  <table id="unused">
    <thead>
      <tr>
        <th>Variable</th>
        <th>Configurations</th>
      </tr>
    </thead>
    <tbody>
    """

    __table_detailed_sig_desc = """
  <table id="sig_desc">
    <thead>
      <tr>
        <th>Unit</th>
        <th>Type</th>
        <th>Class</th>
        <th>Min</th>
        <th>Max</th>
        <th>Lsb</th>
        <th>Offset</th>
        <th>Configs</th>
      </tr>
    </thead>
    <tbody>
"""

    __table_detailed_sig_desc_ext = """
  <table id="sig_desc">
    <thead>
      <tr>
        <th>Unit</th>
        <th>Type</th>
        <th>Init Value</th>
        <th>Min</th>
        <th>Max</th>
        <th>Config Doc</th>
      </tr>
    </thead>
    <tbody>
"""

    __table_detailed_sig_def = """
  <table id="detailed_sig_def">
    <thead>
      <tr>
        <th>Defined in unit</th>
        <th>in projects</th>
      </tr>
    </thead>
    <tbody>
"""

    __table_detailed_sig_cons = """
  <table id="detailed_sig_cons">
    <thead>
      <tr>
        <th>Consumed by units</th>
        <th>in projects</th>
      </tr>
    </thead>
    <tbody>
"""

    def __init__(self, build_cfg, unit_cfgs, sig_ifs):
        """Initialize class instance.

        Args:
            build_cfg (BuildProjConfig): holds all the build configs
            unit_cfg (UnitConfigs): object holding all unit interfaces
            sig_if (CsvSignalInterfaces): object holding signal interface information
        """
        super().__init__('Signal interface report')

        self._build_cfg = build_cfg
        self._unit_cfg_all = unit_cfgs
        self._sig_if_all = sig_ifs

        self._format_unit_data()

    def _gen_sigs_details(self):
        """Generate detailed signal information."""
        self._out_all_sd = {}

        for prj, ifin in self._if_all.items():
            out = ''
            self._if = ifin
            self._sig_if = self._sig_if_all.get(prj)
            out += '  <h2 id="signal_details">Detailed Signal Information</h2>\n'
            for signal in sorted(self._if.keys()):
                out += self._gen_sig_details(signal, prj)
            self._out_all_sd.update({prj: out})

        return self._out_all_sd

    def _gen_sig_details(self, signal, prj):
        """Generate detailed signal information."""
        out = f'  <h3 id="{signal}">{signal}</h3>\n'
        try:
            unit = tuple(self._if[signal]['def'].keys())[0]  # ugly! use the first defining unit.

            # Check that it is a unit, and not from the interface definition
            self._unit_cfg = self._unit_cfg_all.get(prj)
            if unit[:2] == 'Vc':
                per_unit_config = self._unit_cfg.get_per_unit_cfg('all')[unit]['outports'][signal]
                sig_def = defaultdict(lambda: '-', per_unit_config)
                out += f"  <p>{sig_def['description']}</p>"
                out += self.__table_detailed_sig_desc
                out += "      <tr>\n"
                out += f"        <td>{sig_def['unit']}</td>\n"
                out += f"        <td>{sig_def['type']}</td>\n"
                out += f"        <td>{sig_def['class']}</td>\n"
                out += f"        <td>{sig_def['min']}</td>\n"
                out += f"        <td>{sig_def['max']}</td>\n"
                out += f"        <td>{sig_def['lsb']}</td>\n"
                out += f"        <td>{sig_def['offset']}</td>\n"
                out += f"        <td>{sig_def['configs']}</td>\n"
                out += "      </tr>\n"
                out += "    </tbody>\n"
                out += "  </table>\n  <p></p>\n"
            else:
                sig_def = defaultdict(lambda: '-', self._sig_if.get_raw_io_cnfg()[unit][signal])
                out += f"  <p>{sig_def['description']}</p>"
                out += self.__table_detailed_sig_desc_ext
                out += "      <tr>\n"
                out += f"        <td>{sig_def['unit']}</td>\n"
                out += f"        <td>{sig_def['type']}</td>\n"
                out += f"        <td>{sig_def['init']}</td>\n"
                out += f"        <td>{sig_def['min']}</td>\n"
                out += f"        <td>{sig_def['max']}</td>\n"
                out += f"        <td>{unit}</td>\n"
                out += "      </tr>\n"
                out += "    </tbody>\n"
                out += "  </table>\n  <p></p>\n"
        except KeyError:
            pass

        out += self.__table_detailed_sig_def
        prj_str = ""
        unit_str = ""
        for cur_prj, ifin in self._if_all.items():
            self._if = ifin
            prj_str += f"  {cur_prj}"
            try:
                unit_str = ""
                for unit in self._if[signal]['def']:
                    unit_str += '  '.join(unit)
            except KeyError:
                continue
        out += "      <tr>\n"
        out += f"        <td>{unit_str}</td>\n"
        out += f"        <td>{prj_str}</td>\n"
        out += "      </tr>\n"
        out += "    </tbody>\n"
        out += "  </table>\n"

        out += self.__table_detailed_sig_cons
        prj_str = ""
        units_str = ""
        for cur_prj, ifin in self._if_all.items():
            self._if = ifin
            try:
                if 'consumers' in self._if[signal]:
                    prj_str += f"  {cur_prj}"
                    out += "  <p></p>\n"
                    units_str = ""
                    for units in self._if[signal]['consumers']:
                        units_str += '  '.join(units)
            except KeyError:
                continue
        out += "      <tr>\n"
        out += f"        <td>{units_str}</td>\n"
        out += f"        <td>{prj_str}</td>\n"
        out += "      </tr>\n"
        out += "    </tbody>\n"
        out += "  </table>\n"
        return out

    def _format_unit_data(self):
        """Format data per unit."""
        self._if_all = {}
        for prj, unit_cfg in self._unit_cfg_all.items():
            self._if = {}
            self._unit_cfg = unit_cfg
            prj_cfg = self._unit_cfg.get_per_cfg_unit_cfg(prj)
            sig_if = self._sig_if_all.get(prj)
            self._sig_if = sig_if

            # internal signals
            ext_out_prj = self._sig_if.get_external_outputs(prj)
            if 'outports' in prj_cfg:
                for sig, data in prj_cfg['outports'].items():
                    prod_unit = list(data.keys())[0]
                    self._if.setdefault(sig, {}).setdefault('def', {}).setdefault(prod_unit, []).append(prj)
                    cons_units = []
                    for sig_type, osig in ext_out_prj.items():
                        if sig in osig:
                            cons_units = [sig_type]
                            break
                    if sig in prj_cfg['inports']:
                        cons_units.extend(list(prj_cfg['inports'][sig].keys()))
                    if cons_units:
                        self._if[sig].setdefault('consumers', {}).setdefault(tuple(sorted(cons_units)), []).append(prj)

            # external signals
            ext_in_prj = self._sig_if.get_external_inputs(prj)
            for sig_type, osig in ext_in_prj.items():
                for sig, data in osig.items():
                    self._if.setdefault(sig, {}).setdefault('def', {}).setdefault(sig_type, []).append(prj)
                    if 'inports' in prj_cfg and sig in prj_cfg['inports']:
                        (self._if[sig].setdefault('consumers', {})
                                      .setdefault(tuple(sorted(prj_cfg['inports'][sig].keys())), [])
                                      .append(prj))
            self._if_all.update({prj: self._if})
        return self._if_all

    def _gen_signal_toc(self):
        """Generate a signal TOC for the unit with signal inconsistencies.

        Hyperlinks to more in depth signal information.
        """
        self._out_all_st = {}
        for prj, ifin in self._if_all.items():
            out = ''
            self._if = ifin
            out += '  <h2 id="signal_index">Signal index</h2>\n'
            for signal in sorted(self._if.keys()):
                out += f'  <div><a href="#{signal}">{signal}</a></div>\n'
            self._out_all_st.update({prj: out})

        return self._out_all_st

    def _gen_contents_intro(self):
        """Generate report contents from the signal interfaces data.

        Specialises HtmlReport.gen_contents()
        """
        html = []
        html += self.__intro
        return ''.join(html)

    def _gen_contents_signal_info(self):
        """Generate report contents from the signal interfaces data.

        Specialises HtmlReport.gen_contents()
        """
        html = []
        html += self.__signal_info
        return ''.join(html)

    def _generate_report_string_intro(self):
        """Generate a html report as string."""
        html = []
        html += self._gen_header()
        html += self._gen_contents_intro()
        html += self._gen_end()
        return ''.join(html)

    def generate_report_file_if(self, filename):
        """Generate a html report and save to file."""
        filename_intro = filename + '_intro.html'
        with open(filename_intro, 'w', encoding="utf-8") as fhndl:
            fhndl.write(self._generate_report_string_intro())

        self._gen_sigs_details()
        self._gen_signal_toc()
        for prj, out in self._out_all_sd.items():
            html = []
            html += self._gen_header()
            html += self._gen_contents_signal_info()
            html += out
            html += self._out_all_st.get(prj)
            html += self._gen_end()
            with open(f'{filename}_{prj}.html', 'w', encoding="utf-8") as fhndl:
                fhndl.write(''.join(html))



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Environment compatibility check."""

import sys


def check_python_string(python_lower, python_upper=None):
    """Ensure current Python interpreter is a version between lower and upper.

    Arguments:
        python_lower (str): Required lower bound for Python version.
        python_upper (str): Optional upper bound for Python version.
    Raises:
        RuntimeError: If current Python executable is not compatible with powertrain_build.

    """
    versions = [_split_version(python_lower)]
    if python_upper:
        versions.append(_split_version(python_upper))
    check_python_tuple(*versions)


def check_python_tuple(python_lower, python_upper=None):
    """Ensure current Python interpreter is a version between lower and upper.

    Arguments:
        python_lower (2-tuple): Required lower bound for Python version.
        python_upper (2-tuple): Optional upper bound for Python version.
    Raises:
        RuntimeError: If current Python executable is not compatible with powertrain_build.

    """
    cur_version = sys.version_info[:2]

    if cur_version[0] < python_lower[0] or cur_version[1] < python_lower[1]:
        raise RuntimeError(_format_error(f'must be higher than {python_lower}'))

    if python_upper and (cur_version[0] > python_upper[0] or cur_version[1] > python_upper[1]):
        raise RuntimeError(_format_error(f'must be lower than {python_upper}'))


def _split_version(version_string):
    """Split a major.minor style string and returns a 2-tuple of integers."""
    parts = version_string.split('.')
    return (int(parts[0]), int(parts[1]))


def _format_error(message):
    """Return a version error string including current interpreter, version and a custom message."""
    return f'Unsupported Python version ({sys.version_info}), {message}. Path: {sys.executable}'



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Common functions and properties for types."""

TL_TYPES = dict(
    Float32=dict(
        size=4,
        a2l_type='FLOAT32_IEEE',
        a2l_range=('-1e38', '1e38')
    ),
    UInt32=dict(
        size=4,
        a2l_type='ULONG',
        a2l_range=(0, 4294967295)
    ),
    Int32=dict(
        size=4,
        a2l_type='SLONG',
        a2l_range=(-2147483648, 2147483647)
    ),
    UInt16=dict(
        size=2,
        a2l_type='UWORD',
        a2l_range=(0, 65535)
    ),
    Int16=dict(
        size=2,
        a2l_type='SWORD',
        a2l_range=(-32768, 32767)
    ),
    UInt8=dict(
        size=1,
        a2l_type='UBYTE',
        a2l_range=(0, 255)
    ),
    Int8=dict(
        size=1,
        a2l_type='SBYTE',
        a2l_range=(-128, 127)
    ),
    Bool=dict(
        size=1,
        a2l_type='UBYTE',
        a2l_range=(0, 1),
        bitmask='0x01'
    )
)

EC_TYPES = dict(
    real64_T=dict(
        size=8,
        a2l_type='FLOAT64_IEEE',
        a2l_ranger=('-1.8e308', '1.8e308')
    ),
    real_T=dict(
        size=8,
        a2l_type='FLOAT64_IEEE',
        a2l_ranger=('-1.8e308', '1.8e308')
    ),
    time_T=dict(
        size=8,
        a2l_type='FLOAT64_IEEE',
        a2l_ranger=('-1.8e308', '1.8e308')
    ),
    ulong_T=dict(
        size=4,
        a2l_type='ULONG',
        a2l_range=(0, 4294967295)
    ),
    real32_T=dict(
        size=4,
        a2l_type='FLOAT32_IEEE',
        a2l_range=('-1e38', '1e38')
    ),
    uint32_T=dict(
        size=4,
        a2l_type='ULONG',
        a2l_range=(0, 4294967295)
    ),
    uint_T=dict(
        size=4,
        a2l_type='ULONG',
        a2l_range=(0, 4294967295)
    ),
    int32_T=dict(
        size=4,
        a2l_type='SLONG',
        a2l_range=(-2147483648, 2147483647)
    ),
    int_T=dict(
        size=4,
        a2l_type='SLONG',
        a2l_range=(-2147483648, 2147483647)
    ),
    uint16_T=dict(
        size=2,
        a2l_type='UWORD',
        a2l_range=(0, 65535)
    ),
    int16_T=dict(
        size=2,
        a2l_type='SWORD',
        a2l_range=(-32768, 32767)
    ),
    uint8_T=dict(
        size=1,
        a2l_type='UBYTE',
        a2l_range=(0, 255)
    ),
    int8_T=dict(
        size=1,
        a2l_type='SBYTE',
        a2l_range=(-128, 127)
    ),
    boolean_T=dict(
        size=1,
        a2l_type='UBYTE',
        a2l_range=(0, 1),
        bitmask='0x01'
    ),
    uchar_T=dict(
        size=1,
        a2l_type='UBYTE',
        a2l_range=(0, 255)
    ),
    char_T=dict(
        size=1,
        a2l_type='SBYTE',
        a2l_range=(-128, 127)
    ),
    byte_T=dict(
        size=1,
        a2l_type='SBYTE',
        a2l_range=(-128, 127)
    )
)


def _validate_c_type(c_type):
    """Make sure given c-type is handled."""
    if c_type in TL_TYPES:
        return TL_TYPES[c_type]
    if c_type in EC_TYPES:
        return EC_TYPES[c_type]
    raise KeyError(f'Invalid data type: {c_type}')


def byte_size(c_type):
    """Get byte size of a c-type as an int."""
    type_dict = _validate_c_type(c_type)
    return type_dict['size']


def byte_size_string(c_type):
    """Get byte size of a c-type as a string."""
    return str(byte_size(c_type))


def a2l_type(c_type):
    """Get a2l-type of a c-type."""
    type_dict = _validate_c_type(c_type)
    return type_dict['a2l_type']


def a2l_range(c_type):
    """Get a2l-range of a c-type."""
    # TODO: Find out why we default to float32 in a2l.py
    # Probably due to vc_dummy.
    if c_type == 'float32' or c_type is None:
        # TODO Assuming TL for now
        return TL_TYPES['Float32']['a2l_range']
    type_dict = _validate_c_type(c_type)
    return type_dict['a2l_range']


def get_bitmask(c_type):
    """Get bit mask of a c-type.

    Defaults to None for types other than Bool.
    """
    type_dict = _validate_c_type(c_type)
    return type_dict.get('bitmask')


def get_ec_type(tl_type):
    """Get equivalent EC type for a given TL type.

    Args:
        tl_type (str): TL type

    Returns:
        str: EC type
    """
    tl_ec = dict(
        Float32='real32_T',
        UInt32='uint32_T',
        Int32='int32_T',
        UInt16='uint16_T',
        Int16='int16_T',
        UInt8='uint8_T',
        Int8='int8_T',
        Bool='boolean_T'
    )
    return tl_ec[tl_type]


def get_float32_types():
    """Get all EC types that are equivalent to Float32.

    Returns:
        set: EC types
    """
    float32_types = {'Float32'}
    for ec_type, ec_type_def in EC_TYPES.items():
        if ec_type_def == TL_TYPES['Float32']:
            float32_types.add(ec_type)
    return float32_types



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

# -*- coding: utf-8 -*-
"""This module is used to parse the core definition files.

Also provides methods to filter the definitions per project.
"""

import os
import time
from collections import defaultdict
from pathlib import Path
from ruamel.yaml import YAML
from powertrain_build import build_defs
from powertrain_build.xlrd_csv import WorkBook
from powertrain_build.problem_logger import ProblemLogger


class Core(ProblemLogger):
    """A class holding core configuration data."""

    __wrk_sheets = {'EventIDs': {'mdl_col': 0,
                                 'id_col': 1,
                                 'desc_col': 2,
                                 'fid_col': 3,
                                 'com_col': 4,
                                 'data_col': 5},
                    'FunctionIDs': {'mdl_col': 0,
                                    'id_col': 1,
                                    'desc_col': 2,
                                    'com_col': 3,
                                    'data_col': 4},
                    'IUMPR': {'mdl_col': 0,
                              'id_col': 1,
                              'desc_col': 2,
                              'fid_col': 3,
                              'com_col': 4,
                              'data_col': 5},
                    'Mode$06': {'mdl_col': 0,
                                'id_col': 1,
                                'desc_col': 2,
                                'fid_col': 3,
                                'UAS_col': 4,
                                'com_col': 5,
                                'data_col': 6},
                    'Ranking': {'mdl_col': 0,
                                'id_col': 1,
                                'desc_col': 2,
                                'fid_col': 3,
                                'com_col': 4,
                                'data_col': 5}}

    _NoFid = '_NoFid'

    def __init__(self, project_cfg, unit_cfgs):
        """Parse the config files to an internal representation.

        Args:
            project_cfg (BuildProjConfig): Project configuration
            unit_cfgs (UnitConfigs): Unit definitions
        """
        super().__init__()
        self._prj_cfg = project_cfg
        self._unit_cfgs = unit_cfgs
        self._csv_files = None
        self._read_csv_files(self._prj_cfg.get_prj_cfg_dir())
        self._parse_core_config()

    def _read_csv_files(self, config_path):
        """Metod for reading the core csv confgiuration files."""
        self.info('******************************************************')
        self.info('Start parsing the core configuration files')
        start_time = time.time()
        csv_file_names = []
        for sheet in self.__wrk_sheets:
            fname = os.path.join(config_path,
                                 'CoreIdNameDefinition_' + sheet + '.csv')
            self.debug('Core csv: %s', fname)
            csv_file_names.append(fname)
        self._csv_files = WorkBook(csv_file_names)
        self.info('Finished parsing the core configuration files (in %4.2f s)', time.time() - start_time)

    def _parse_core_config(self):
        """Parse core IDs for all projects."""
        # Parse sheet by sheet
        tmp_ids = {}
        for sheet_name, cols in self.__wrk_sheets.items():
            worksheet = self._csv_files.sheet_by_name(sheet_name)
            prj_row = worksheet.row(1)
            curr_sheet = tmp_ids[sheet_name] = {}
            for prj_col in range(cols['data_col'], len(prj_row)):
                prj = prj_row[prj_col].value
                curr_id = curr_sheet[prj] = {}
                for curr_row in range(2, worksheet.nrows):
                    row = worksheet.row(curr_row)
                    val = row[prj_col].value
                    if val and val.strip() in 'xX':
                        func_tmp = row[cols['id_col']].value.strip()
                        if func_tmp != '':
                            if sheet_name == 'Mode$06':
                                uas_val = row[cols['UAS_col']].value
                                if isinstance(uas_val, float):
                                    uas_val = int(uas_val)
                                curr_id[func_tmp] = (row[cols['desc_col']].value,
                                                     str(uas_val))
                            else:
                                curr_id[func_tmp] = row[
                                    cols['desc_col']].value
        # Reformat from Sheet->Proj->Id to Proj->Sheet->Id
        self._ids = defaultdict(dict)
        for s_name, s_data in tmp_ids.items():
            for p_name, p_data in s_data.items():
                self._ids[p_name][s_name] = p_data

    def get_core_ids_proj(self, project):
        """Get the core IDs for a project.

        Returns:
            dict: Core IDs

        """
        core_ids = self._ids[project]
        # Check for unused core symbols
        for sheet_name, sheet_data in core_ids.items():
            for id_ in sheet_data:
                if not self._unit_cfgs.check_if_in_per_cfg_unit_cfg('core', id_):
                    self.debug('Core symbol not used in current project: %s/%s', sheet_name, id_)
        # Check for undefined core symbols in unit configs
        ucfg = self._unit_cfgs.get_per_cfg_unit_cfg()
        for id_ in ucfg.get('core', {}):
            for _, core_sheet in core_ids.items():
                if id_ in core_sheet:
                    break
            else:
                self.warning('Core symbol not defined for current project: %s', id_)
        return core_ids

    def get_current_core_config(self):
        """Return all the core configuration parameters for the current project.

        Returns:
            dict: All the core configuration parameters

        """
        return self.get_core_ids_proj(self._prj_cfg.get_prj_config())


class HICore(ProblemLogger):
    """A class holding HI core configuration data."""

    DTC_DEFINITION_FILE_NAME = 'DTCs.yaml'
    FILE_NAME = 'VcCoreSupplierAbstraction'

    def __init__(self, project_cfg, unit_cfgs):
        """Parse the config files to an internal representation.

        Args:
            project_cfg (BuildProjConfig): Project configuration.
            unit_cfgs (UnitConfigs): Unit definitions.
        """
        super().__init__()
        self._prj_cfg = project_cfg
        self._unit_cfgs = unit_cfgs
        self.diagnostic_trouble_codes = self.get_diagnostic_trouble_codes()

    def _get_project_dtcs(self):
        """Return a set with DTCs in the currently included SW-Units.

        Returns:
            (set): Set of DTCs in the currently included SW-Units.
        """
        project_dtcs = set()
        unit_cfg = self._unit_cfgs.get_per_unit_cfg()
        for unit, data in unit_cfg.items():
            event_data = data.get('core', {}).get('Events')
            if event_data is None:
                self.critical(f'No "core" or "core.Events" key in unit config for {unit}.')
                continue
            project_dtcs |= set(event_data.keys())
        return project_dtcs

    def _read_dtc_yaml_file(self):
        """Return a set with DTCs loaded from the project DTC yaml file.

        Returns:
            (set): Set of DTCs loaded from the DTC yaml file.
        """
        diagnostic_trouble_codes = {}
        dtc_definition_file_path = Path(self._prj_cfg.get_prj_cfg_dir(),  self.DTC_DEFINITION_FILE_NAME)
        if dtc_definition_file_path.exists():
            with dtc_definition_file_path.open(mode='r', encoding='utf-8') as dtc_fh:
                yaml = YAML(typ='safe', pure=True)
                diagnostic_trouble_codes = yaml.load(dtc_fh)
        else:
            self.warning(
                'Unable to generate DTC function calls. '
                f'Cannot find file: {dtc_definition_file_path.as_posix()}.'
            )
        return diagnostic_trouble_codes

    def get_diagnostic_trouble_codes(self):
        """Return a set of DTCs appearing in both the project and the project yaml file.

        Returns:
            (dict): Dict of DTCs, project yaml dict where the keys also appear in the project.
        """
        project_dtcs = self._get_project_dtcs()
        yaml_dtc_dict = self._read_dtc_yaml_file()
        yaml_dtcs = set(yaml_dtc_dict.keys())
        dtcs_not_in_project = yaml_dtcs - project_dtcs
        dtcs_not_in_yaml = project_dtcs - yaml_dtcs
        for key in dtcs_not_in_project:
            self.warning(f'Ignoring DTC {key} since it does not appear in any model.')
            del yaml_dtc_dict[key]
        for key in dtcs_not_in_yaml:
            self.warning(f'Ignoring DTC {key} since it does not appear in the project DTC yaml file.')
        return yaml_dtc_dict

    def get_header_content(self):
        """Get content for the DTC header file.

        Returns:
            (list(str)): List of lines to write to the DTC header file.
        """
        name = self._prj_cfg.get_a2l_cfg()['name']
        header_guard = f'{self.FILE_NAME.upper()}_H'
        header = [
            f'#ifndef {header_guard}\n',
            f'#define {header_guard}\n',
            '\n',
            '/* Core API Supplier Abstraction */\n',
            '\n',
            '#include "tl_basetypes.h"\n',
            f'#include "Rte_{name}.h"\n',
            '\n'
        ]
        footer = [f'\n#endif /* {header_guard} */\n']

        if not self.diagnostic_trouble_codes:
            return header + footer

        body = [
            '/* enum EventStatus {passed=0, failed=1, prepassed=2, prefailed=3} */\n',
            '#define Dem_SetEventStatus(EventName, EventStatus)',
            '  ',
            f'{self.FILE_NAME}_##EventName##_SetEventStatus(EventStatus)\n'
        ]
        body.append(f'\n#include "{build_defs.PREDECL_CODE_ASIL_D_START}"\n')
        for dtc_name in self.diagnostic_trouble_codes:
            body.append(f'UInt8 {self.FILE_NAME}_{dtc_name}_SetEventStatus(UInt8 EventStatus);\n')
        body.append(f'#include "{build_defs.PREDECL_CODE_ASIL_D_END}"\n')
        return header + body + footer

    def get_source_content(self):
        """Get content for the DTC source file.

        Returns:
            (list(str)): List of lines to write to the DTC source file.
        """
        header = [
            f'#include "{self.FILE_NAME}.h"\n',
            '\n'
        ]

        if not self.diagnostic_trouble_codes:
            return header

        body = [f'#include "{build_defs.CVC_CODE_ASIL_D_START}"\n']
        for dtc_name, dtc_id in self.diagnostic_trouble_codes.items():
            # hex function removes leading 0, below solution zero pads to 6 digits
            dtc_hex_str = f"0x{dtc_id:06X}"
            body.extend([
                f'UInt8 {self.FILE_NAME}_{dtc_name}_SetEventStatus(UInt8 EventStatus)\n',
                '{\n',
                f'    Rte_Call_Event_DTC_{dtc_hex_str}_SetEventStatus(EventStatus);\n',
                '    return 0;\n',
                '}\n',
                '\n'
            ])
        body.append(f'#include "{build_defs.CVC_CODE_ASIL_D_END}"\n')

        return header + body

    def generate_dtc_files(self):
        """Generate required HI Core header files.
        Only use for some projects, which doesn't copy static code."""
        file_contents = {
            '.h': self.get_header_content(),
            '.c': self.get_source_content()
        }
        src_dst_dir = self._prj_cfg.get_src_code_dst_dir()
        for extension, content in file_contents.items():
            file_path = Path(src_dst_dir, self.FILE_NAME + extension)
            with file_path.open(mode='w', encoding='utf-8') as file_handler:
                file_handler.writelines(content)


class ZCCore(ProblemLogger):
    """A class holding ZC core configuration data."""

    FILE_NAME = 'VcCoreSupplierAbstraction'

    def __init__(self, project_cfg, unit_cfgs):
        """Parse the config files to an internal representation.

        Args:
            project_cfg (BuildProjConfig): Project configuration.
            unit_cfgs (UnitConfigs): Unit definitions.
        """
        super().__init__()
        self._prj_cfg = project_cfg
        self._unit_cfgs = unit_cfgs
        self.project_dtcs = self._get_project_dtcs()

    def _get_project_dtcs(self):
        """Return a set with DTCs in the currently included SW-Units.

        Returns:
            (set): Set of DTCs in the currently included SW-Units.
        """
        project_dtcs = set()
        unit_cfg = self._unit_cfgs.get_per_unit_cfg()
        for unit, data in unit_cfg.items():
            event_data = data.get('core', {}).get('Events')
            if event_data is None:
                self.critical(f'No "core" or "core.Events" key in unit config for {unit}.')
                continue
            project_dtcs |= set(event_data.keys())
        return project_dtcs

    def get_diagnostic_trouble_codes(self, event_data):
        """Return a set of DTCs appearing in both the project and the project yaml file.

        Args:
            event_data (dict): Diagnositc event data.
        Returns:
            (dict): Dict of DTCs, project yaml dict where the keys also appear in the project.
        """
        valid_dtcs = {}

        dtcs_not_in_yaml = self.project_dtcs - set(event_data.keys())
        for key in dtcs_not_in_yaml:
            self.warning(f'Ignoring DTC {key} since it does not appear in the project diagnostics yaml file.')

        for dtc_name, dtc_data in event_data.items():
            if dtc_name not in self.project_dtcs and not dtc_data.get("manual", False):
                self.warning(f'Ignoring DTC {dtc_name}, not defined in any model.')
                continue
            valid_dtcs[dtc_name] = {k: v for k, v in dtc_data.items() if k != "manual"}
        return valid_dtcs

    def get_header_content(self):
        """Get content for the DTC header file.

        Returns:
            (list(str)): List of lines to write to the DTC header file.
        """
        name = self._prj_cfg.get_composition_config("softwareComponentName")
        header_guard = f'{self.FILE_NAME.upper()}_H'
        header = [
            f'#ifndef {header_guard}\n',
            f'#define {header_guard}\n',
            '\n',
            '/* Core API Supplier Abstraction */\n',
            '\n',
            '#include "tl_basetypes.h"\n',
            f'#include "Rte_{name}.h"\n',
            '\n'
        ]
        footer = [f'\n#endif /* {header_guard} */\n']

        body = [
            '/* EventStatus is an enumeration with members {passed=0, failed=1, prepassed=2, prefailed=3} */\n',
            '#define Dem_SetEventStatus(EventName, EventStatus)',
            '  ',
            'Rte_Call_Event_##EventName##_SetEventStatus(EventStatus)\n'
        ]
        return header + body + footer

    def generate_dtc_files(self):
        """Generate required ZC Core header files.
        Only use for some projects, which doesn't copy static code."""
        includeDiagnostics = self._prj_cfg.get_composition_config("includeDiagnostics")
        if includeDiagnostics in ["manual", "manual_dtcs"]:
            self.warning(f'includeDiagnostics is set to {includeDiagnostics}, not generating DTC files.')
            return
        file_contents = {
            '.h': self.get_header_content()
        }
        src_dst_dir = self._prj_cfg.get_src_code_dst_dir()
        for extension, content in file_contents.items():
            file_path = Path(src_dst_dir, self.FILE_NAME + extension)
            with file_path.open(mode='w', encoding='utf-8') as file_handler:
                file_handler.writelines(content)



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Module with generation utils."""

import argparse
from ruamel.yaml import YAML
from pathlib import Path
from powertrain_build.interface.application import Application, get_internal_domain
from powertrain_build.interface.base import filter_signals
from powertrain_build.lib import logger

LOGGER = logger.create_logger("CSP interface generation utils")


def add_base_args(parser: argparse.ArgumentParser):
    """ Base parser that adds config argument.

    Returns:
        parser (ArgumentParser): Base parser
    """
    parser.add_argument("config", help="The project configuration file", type=Path)


def get_client_name(args, app):
    """ Get client name for app.

    Args:
        app (Application): Parsed project configuration
    Returns:
        name (str): Name of the project
    """
    return args.client_name if args.client_name else app.name


def process_app(config):
    """ Get an app specification for the current project

    Entrypoint for external scripts.

    Args:
        config (pathlib.Path): Path to the ProjectCfg.json
    Returns:
        app (Application): powertrain-build project
    """
    app = Application()
    app.parse_definition(config)
    return app


def get_interface(app, interface_type):
    """Get interface(hal/dp/zc/service) to application

    Args:
        app (Application): Pybuild project
        interface_type (BaseApplication): A type of interface
    Returns:
        spec (obj): obj for hal/dp/zc/service class
    """
    spec = {}
    rasters = app.get_rasters()
    LOGGER.debug("Rasters: %s", rasters)
    translation_files = app.get_translation_files()
    interface_type.parse_definition(translation_files)
    internal = get_internal_domain(rasters)
    properties_from_json = [
        {"destination": "min", "source": "min", "default": "-"},
        {"destination": "max", "source": "max", "default": "-"},
        {"destination": "variable_type", "source": "type"},
        {"destination": "offset", "source": "offset", "default": "-"},
        {"destination": "factor", "source": "lsb", "default": 1},
        {"destination": "description", "source": "description"},
        {"destination": "unit", "source": "unit", "default": "-"},
    ]
    # TODO We read all yaml files at once, should we "add_signals" for all rasters at once?
    # For example, service.check_endpoints will miss endpoint mismatches that reside in different rasters.
    for raster in rasters:
        interface_type.name = raster.name
        interface_type.clear_signal_names()
        interface_type.add_signals(
            filter_signals(raster.insignals, internal), "insignals", properties_from_json)
        interface_type.add_signals(raster.outsignals, "outsignals", properties_from_json)
        LOGGER.debug("Current communication interface: %s", interface_type)
        spec[raster.name] = interface_type.to_dict()
    return spec


def get_method_interface(app):
    """ Get method interface
    Args:
        app (Application): Application
    Returns:
        spec (dict): Specification for csp methods
    """
    spec = {}

    for method in app.get_methods():
        method_spec = {}
        method_spec['name'] = method.name
        method_spec['primitive'] = method.get_primitive(method.name)
        method_spec['namespace'] = method.namespace
        if method.description:
            method_spec['description'] = method.description
        inports = {}
        outports = {}
        for signal in method.signals:
            signal_spec = {
                'primitive': method.get_primitive(signal.name),
                'type': signal.properties['type'],
                'variable_type': signal.properties['type'],
                'variable': signal.name,
                'range': signal.properties['range'],
                'init': 0,
                'description': '',
                'unit': ''
            }
            if signal in method.outsignals:
                outports[signal.name] = signal_spec
            else:
                inports[signal.name] = signal_spec

        ports = {'in': inports, 'out': outports}
        method_spec['ports'] = ports
        spec[method.name] = method_spec
    return spec


def write_to_file(content, output, is_yaml=False):
    """ Write to cmake.

    Args:
        content (str): cmake
        output (Path): File to write.
        yaml (bool): Dump yaml
    """
    output.parent.mkdir(parents=True, exist_ok=True)
    with open(output, "w", encoding="utf-8") as file_handler:
        if is_yaml:
            yaml = YAML()
            yaml.dump(content, file_handler)
        else:
            file_handler.write(content)



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Python module used for reading device proxy arxml:s"""
from ruamel.yaml import YAML
import enum
from powertrain_build.interface.base import BaseApplication, Signal
from powertrain_build.lib import logger

LOGGER = logger.create_logger("device_proxy")


class MissingDevice(Exception):
    """Exception to raise when device is missing"""
    def __init__(self, dp):
        self.message = f"Device proxy {dp} missing from deviceDomains.json"


class BadYamlFormat(Exception):
    """Exception to raise when in/out signal is not defined."""
    def __init__(self, message):
        self.message = message


class DPAL(BaseApplication):
    """Device Proxy abstraction layer"""

    dp_position = enum.Enum(
        "Position",
        names=[
            "domain",
            "property",
            "variable_type",
            "property_interface_type",
            "property_manifest_type",
            "offset",
            "factor",
            "default",
            "length",
            "min",
            "max",
            "enum",
            "init",
            "description",
            "unit",
            "group",
            "strategy",
            "debug",
            "dependability",
            "port_name"
        ],
    )

    def __repr__(self):
        """String representation of DPAL"""
        return (
            f"<DPAL {self.name}"
            f" app_side insignals: {len(self.signal_names['other']['insignals'])}"
            f" app_side outsignals: {len(self.signal_names['other']['outsignals'])}>"
        )

    def __init__(self, base_application):
        """Create the interface object

        Args:
            base_application (BaseApplication): Primary object of an interface
                                                Usually a raster, but can be an application or a model too.
        """
        self.name = ""
        self.dp_translations = {}
        # We do not care about domain when looking from a project perspective,
        # we only care when generating manifests for csp.
        self.domain_filter = None
        self.signal_names = {
            "dp": {"insignals": set(), "outsignals": set()},
            "other": {"insignals": set(), "outsignals": set()},
        }
        self.e2e_sts_signals = set()
        self.base_application = base_application
        self.translations_files = []
        self.device_domain = base_application.get_domain_mapping()
        self.signal_primitives_list = []

    def clear_signal_names(self):
        """Clear signal names

        Clears defined signal names (but not signal properties).
        """
        self.signal_names = {
            "dp": {"insignals": set(), "outsignals": set()},
            "other": {"insignals": set(), "outsignals": set()},
        }

    def add_signals(self, signals, signal_type="insignal", properties=[]):
        """Add signal names and properties

        Args:
            signals (list(Signals)): Signals to use
            signal_type (str): 'insignals' or 'outsignals'
            properties (list(str)): signal definition properties, default = []
        """
        opposite = {"insignals": "outsignals", "outsignals": "insignals"}
        dp_type = opposite[signal_type]
        for signal in signals:
            LOGGER.debug("Adding signal: %s", signal)
            temp_set = set()
            for translation in self.dp_translations.get(signal.name, []):
                temp_list = list(translation)
                domain = translation[self.dp_position.domain.value]
                group = translation[self.dp_position.group.value]
                dp_signal = translation[self.dp_position.property.value]
                self.check_signal_property(domain, group, dp_signal, signal_type)
                self.signal_names["dp"][dp_type].add(dp_signal)
                for enum_property in properties:
                    LOGGER.debug("Modifying property: %s", enum_property)
                    value = signal.properties[enum_property["source"]]
                    if value == "-":
                        value = enum_property["default"]
                    temp_list[
                        self.dp_position[enum_property["destination"]].value
                    ] = value
                temp_set.add(tuple(temp_list))
                self.dp_translations[signal.name] = temp_set
            self.signal_names["other"][signal_type].add(signal.name)
        for e2e_sts_signal_name in self.e2e_sts_signals:
            if e2e_sts_signal_name not in self.signal_names["other"]["insignals"]:
                LOGGER.warning("E2E check signal %s not used in any model.", e2e_sts_signal_name)
                self.signal_names["other"][signal_type].add(e2e_sts_signal_name)
        self.check_groups()
        LOGGER.debug("Registered signal names: %s", self.signal_names)

    def check_signal_property(self, domain, group, property_name, signal_type):
        """Check if we have only one signal written for the same property.

        Args:
            domain (str): signal domain
            group (str): signal group
            property_name (str): signal property
            signal_type (str): 'insignals' or 'outsignals'
        """
        primitive_value = ""
        for value in [domain, group, property_name]:
            if value:
                if primitive_value == "":
                    primitive_value = value
                else:
                    primitive_value = primitive_value + '.' + value
        if primitive_value == "":
            raise Exception("The primitive does not contain any value!")
        directional_primitive = f"{primitive_value}.{signal_type}"
        self.check_property(directional_primitive, signal_type)

    def check_property(self, property_spec, signal_type):
        """Check if we have only one signal written for the same property.

        Args:
            property_spec (str): property specification
            signal_type (str): 'insignals' or 'outsignals'
        """
        if property_spec in self.signal_primitives_list:
            error_msg = (f"You can't write {property_spec} as "
                         f"{signal_type} since this primitive has been used."
                         " Run model_yaml_verification to identify exact models.")
            raise Exception(error_msg)
        self.signal_primitives_list.append(property_spec)

    def check_groups(self):
        """Check and crash if signal group contains both produces and consumes signals."""
        groups = {}
        for signal_name, signal_specs in self.dp_translations.items():
            if signal_name in self.signal_names["other"]['insignals']:
                consumed = True
            elif signal_name in self.signal_names["other"]['outsignals']:
                consumed = False
            else:
                continue
            for signal_spec in signal_specs:
                group = signal_spec[self.dp_position.group.value]
                if group is None:
                    continue
                domain = signal_spec[self.dp_position.domain.value]
                key = (domain, group)
                if key not in groups:
                    groups[key] = {"consumed": consumed,
                                   "signals": set()}
                groups[key]["signals"].add(signal_name)
                assert consumed == groups[key]["consumed"], \
                    f"Signal group {group} for {domain} contains both consumed and produced signals"

    @staticmethod
    def read_translation(translation_file):
        """Read specification of the format:

        service:
          interface:
            properties:
              - endpoint_name:
                - signal: name
                  property: name
                - signal: name
                  property: name
        hal:
          hal_name:
            - primitive_endpoint:
              - insignal: name
          hal_name:
            - struct_endpoint:
              - insignal: name1
                property: member1
              - insignal: name2
                property: member2
          ecm:
            - signal: name
        signals:
          tvrl:
            - signal: name
              property: can_name
              offset: offset
              factor: scaling

        Args:
            translation_file (Path): file with specs

        Returns:
            yaml_data (dict): Loaded YAML data as dict, empty if not found
        """
        if not translation_file.is_file():
            LOGGER.warning("No file found for %s", translation_file)
            return {}
        with open(translation_file, encoding="utf-8") as translation:
            yaml = YAML(typ='safe', pure=True)
            raw = yaml.load(translation)
        return raw

    def parse_group_definitions(self, signal_groups):
        """Parse group definitions.

        Args:
            signal_groups (dict): parsed yaml file.
        """
        for dp_name, group_definitions in signal_groups.items():
            for group in group_definitions:
                port_name = None
                if 'portname' in group:
                    port_name = group.pop('portname')
                for group_name, signals in group.items():
                    self.parse_signal_definitions({dp_name: signals}, group_name, port_name)

    def parse_signal_definitions(self, signals_definition, group=None, port_name=None):
        """Parse signal definitions.

        Args:
            signals_definition (dict): parsed yaml file.
            group (str): Name of signal group, if signal belongs to a group.
            port_name (str): Name of signal port, if there is one.
        """
        enumerations = self.base_application.enumerations
        for dp_name, dp_specification in signals_definition.items():
            for specification in dp_specification:
                in_out_signal = [key for key in specification.keys() if 'signal' in key]
                base_signal = None
                signal_name = None
                if "in" in in_out_signal[0]:
                    for signal in self.base_application.insignals:
                        if signal.name == specification["insignal"]:
                            base_signal = signal
                            signal_name = signal.name
                elif "out" in in_out_signal[0]:
                    for signal in self.base_application.outsignals:
                        if signal.name == specification["outsignal"]:
                            base_signal = signal
                            signal_name = signal.name
                else:
                    raise BadYamlFormat(f"in/out signal for {dp_name} is missing.")
                if base_signal is None:
                    continue
                base_properties = self.base_application.get_signal_properties(
                    base_signal
                )
                if base_properties["type"] in enumerations:
                    underlying_data_type = enumerations[base_properties['type']]['underlying_data_type']
                    interface_type = underlying_data_type
                    manifest_type = underlying_data_type
                    if 'init' not in specification:
                        if enumerations[base_properties['type']]['default_value'] is not None:
                            init_value = enumerations[base_properties['type']]['default_value']
                        else:
                            LOGGER.warning('Initializing enumeration %s to "zero".', base_properties['type'])
                            init_value = [
                                k for k, v in enumerations[base_properties['type']]['members'].items() if v == 0
                            ][0]
                    else:
                        init_value = specification.get("init", 0)
                else:
                    interface_type = base_properties["type"]
                    manifest_type = base_properties["type"]
                    init_value = specification.get("init", 0)

                if "out" in in_out_signal[0] and "strategy" in specification:
                    LOGGER.warning('Cannot set read strategy for outsignal %s, using "Always".', signal_name)
                    strategy = "Always"
                else:
                    strategy = specification.get("strategy", "Always")
                if strategy not in self.read_strategies:
                    LOGGER.warning('Invalid strategy %s, using "Always" instead.', strategy)
                    strategy = "Always"

                if group is not None and specification.get("portname", None) is not None:
                    raise BadYamlFormat(f"Port name should be on group level not signal level: {dp_name}")
                port_name_tmp = port_name if port_name is not None else specification.get("portname", None)

                is_safe_signal = specification.get("dependability", False)

                if signal_name not in self.dp_translations:
                    self.dp_translations[signal_name] = set()
                domain = self._get_domain(dp_name)
                self.dp_translations[signal_name].add(
                    (
                        "enum_0",  # read from this tuple using the dp_position enum. Enum starts at 1 though.
                        domain,
                        specification["property"],
                        specification.get("type"),
                        interface_type,
                        manifest_type,
                        specification.get("offset"),
                        specification.get("factor"),
                        specification.get("default"),
                        specification.get("length"),
                        specification.get("min"),
                        specification.get("max"),
                        specification.get("enum"),
                        init_value,
                        specification.get("description"),
                        specification.get("unit"),
                        group,
                        strategy,
                        specification.get("debug", False),
                        is_safe_signal,
                        port_name_tmp
                    )
                )

                enable_e2e_sts = self.base_application.pybuild['build_cfg'].get_enable_end_to_end_status_signals()
                if enable_e2e_sts and is_safe_signal and group is not None:
                    e2e_sts_property = f"{group}E2eSts"
                    e2e_sts_signal_name = f"sVc{domain}_D_{e2e_sts_property}"

                    if signal_name == e2e_sts_signal_name:
                        raise BadYamlFormat(f"Don't put E2E status signals ({signal_name}) in yaml interface files.")

                    if e2e_sts_signal_name not in self.dp_translations:
                        self.dp_translations[e2e_sts_signal_name] = set()
                        self.dp_translations[e2e_sts_signal_name].add(
                            (
                                "enum_0",  # read from this tuple using the dp_position enum. Enum starts at 1 though.
                                domain,
                                e2e_sts_property,
                                "UInt8",
                                "UInt8",
                                "UInt8",
                                0,
                                1,
                                None,
                                None,
                                0,
                                255,
                                None,
                                255,
                                f"E2E status code for E2E protected signal (group) {signal_name}.",
                                None,
                                group,
                                strategy,
                                False,
                                is_safe_signal,
                                port_name_tmp
                            )
                        )
                        self.e2e_sts_signals.add(e2e_sts_signal_name)

    def parse_definition(self, definition):
        """Parses all definition files

        Args:
            definition (list(Path)): Definition files
        """

        for translation in definition:
            raw = self.read_translation(translation)
            self.parse_signal_definitions(raw.get("signals", {}))
            self.parse_group_definitions(raw.get("signal_groups", {}))

    def get_signal_properties(self, signal):
        """Get signal properties for signal

        Calls self.base_application to get signal properties

        Args:
            signal (Signal): Signal to get properties for
        """
        self.base_application.get_signal_properties(signal)

    def _get_signals(self):
        """Read signals"""
        self.parse_definition(self.translations_files)

    def _get_domain(self, device_proxy):
        """Get domain for device proxy

        Args:
            device_proxy (str): Name of device proxy
        Returns:
            domain (str): Name of the domain
        """
        if device_proxy not in self.device_domain:
            raise MissingDevice(device_proxy)
        return self.device_domain[device_proxy]

    def _allow_domain(self, domain):
        """Check if device proxy is in current domain_filter

        If there is no filter, the device is seen as part of the filter

        Args:
            domain (str): Name of the domain
        Returns:
            filtered (bool): True if device is not filtered away
        """
        return self.domain_filter is None or domain in self.domain_filter

    def get_signals(self, signal_type="insignals"):
        """Get signals to and from a dp abstraction

        If it is set to False, we look at the application side.

        Args:
            signal_type (str): insignals or outsignals
        Returns:
            signals (list): Signals in the interface
        """
        signal_names = self.signal_names["other"][signal_type]

        signals = []
        for name in self._allowed_names(signal_names):
            signals.append(Signal(name, self))
        return signals

    @property
    def insignals(self):
        """ Signals going to the device proxy. """
        return self.get_signals("insignals")

    @property
    def outsignals(self):
        """ Signals sent from the device proxy. """
        return self.get_signals("outsignals")

    def dp_spec_to_dict(self, signal_spec, signal_name):
        """Convert signal specification to dict.

        Args:
            signal_spec (tuple): Signal specification
            signal_name (str): Signal name
        Returns:
            signal_spec (dict): Signal specification
        """
        return {
            "variable": signal_name,
            "variable_type": signal_spec[self.dp_position.variable_type.value],
            "property_type": signal_spec[self.dp_position.property_interface_type.value],
            "domain": signal_spec[self.dp_position.domain.value],
            "default": signal_spec[self.dp_position.default.value],
            "length": signal_spec[self.dp_position.length.value],
            "property": signal_spec[self.dp_position.property.value],
            "offset": signal_spec[self.dp_position.offset.value],
            "factor": signal_spec[self.dp_position.factor.value],
            "range": {
                "min": signal_spec[self.dp_position.min.value],
                "max": signal_spec[self.dp_position.max.value],
            },
            "init": signal_spec[self.dp_position.init.value],
            "description": signal_spec[self.dp_position.description.value],
            "unit": signal_spec[self.dp_position.unit.value],
            "group": signal_spec[self.dp_position.group.value],
            "strategy": signal_spec[self.dp_position.strategy.value],
            "debug": signal_spec[self.dp_position.debug.value],
            "dependability": signal_spec[self.dp_position.dependability.value],
            "port_name": signal_spec[self.dp_position.port_name.value]
        }

    @classmethod
    def dp_spec_for_manifest(cls, signal_spec):
        """Convert signal specification to dict for a signal manifest.

        Args:
            signal_spec (tuple): Signal specification
        Returns:
            signal_spec (dict): Signal specification
        """
        spec = {
            "name": signal_spec[cls.dp_position.property.value],
            "type": signal_spec[cls.dp_position.property_manifest_type.value],
        }
        for key, value in {
            "default": cls.dp_position.default.value,
            "length": cls.dp_position.length.value,
            "enum": cls.dp_position.enum.value,
            "description": cls.dp_position.description.value,
            "unit": cls.dp_position.unit.value,
        }.items():
            if signal_spec[value] is not None:
                spec[key] = signal_spec[value]
        if (
            signal_spec[cls.dp_position.min.value] is not None
            and signal_spec[cls.dp_position.max.value] is not None
            and cls.dp_position.enum.value is not None
        ):
            spec["range"] = {
                "min": signal_spec[cls.dp_position.min.value],
                "max": signal_spec[cls.dp_position.max.value],
            }
        return spec

    def to_dict(self):
        """Method to generate dict to be saved as yaml

        Returns:
            spec (dict): Signalling specification
        """
        spec = {"consumer": [], "producer": []}
        for signal_name, signal_spec in self._allowed_names_and_specifications(
                self.signal_names["other"]["insignals"]):
            spec['consumer'].append(
                self.dp_spec_to_dict(signal_spec, signal_name)
            )
        for signal_name, signal_spec in self._allowed_names_and_specifications(
                self.signal_names["other"]["outsignals"]):
            spec['producer'].append(
                self.dp_spec_to_dict(signal_spec, signal_name)
            )
        return spec

    def to_manifest(self, client_name):
        """Method to generate dict to be saved as yaml
        Args:
            client_name (str): Name of the client in signal comm
        Returns:
            spec (dict): Signal manifest for using a Device proxy
        """
        manifest = {"name": client_name}
        manifest["consumes"] = self.insignals_dp_manifest(client_name)
        manifest["produces"] = self.outsignals_dp_manifest(client_name)
        manifest = self.cleanup_dp_manifest(manifest)
        if "consumes" not in manifest and "produces" not in manifest:
            return None
        return {"signal_info": {"version": 0.2, "clients": [manifest]}}

    def _generator(self, signal_names, unique_names=False):
        """Iterate over signals for allowed devices

        If unique_names is True, the iterator does not yield the same signal twice
        if unique_names is False, it yields each allowed signal spec with the signal name

        Args:
            signal_names (list): allowed signals
        Yields:
            name (str): Name of the signal
            specification (dict): signal specification for allowed device
        """
        for signal_name, specifications in (
                (name, spec) for name, spec in self.dp_translations.items()
                if name in signal_names):
            for specification in (
                    spec for spec in specifications
                    if self._allow_domain(spec[self.dp_position.domain.value])):
                if unique_names:
                    yield signal_name, specification
                    break
                yield signal_name, specification

    def _allowed_names(self, signal_names):
        """ Iterate over signal names for allowed devices

        Args:
            signal_names (list): allowed signals
        Yields:
            name (str): Signal name
        """
        for name, _ in self._generator(signal_names, unique_names=True):
            yield name

    def _allowed_specifications(self, signal_names):
        """ Iterate over signal specifications for allowed devices

        Args:
            signal_names (list): allowed signals
        Yields:
            specification (dict): Specification for a signal for an allowed device
        """
        for _, spec in self._generator(signal_names, unique_names=False):
            yield spec

    def _allowed_names_and_specifications(self, signal_names):
        """ Iterate over signal specifications for allowed devices

        Args:
            signal_names (list): allowed signals
        Yields:
            name (str): Signal name
            specification (dict): Specification for the signal for an allowed device
        """
        for name, spec in self._generator(signal_names, unique_names=False):
            yield name, spec

    def insignals_dp_manifest(self, client_name):
        """ Create consumes part of manifest for reading signals from device proxies
        Args:
            client_name (str): Name of the client in signal comm
        """
        consumes = [{"name": client_name, "signal_groups": []}]
        signal_names = self.signal_names["other"]["insignals"]
        consumed_groups = set()
        for signal_spec in self._allowed_specifications(signal_names):
            group = signal_spec[self.dp_position.group.value]
            if group is not None:
                consumed_groups.add(group)
            else:
                consumes[0]["signal_groups"].append(
                    {"name": signal_spec[self.dp_position.property.value]}
                )
        for group in consumed_groups:
            consumes[0]["signal_groups"].append(
                {"name": group}
            )
        return consumes

    def outsignals_dp_manifest(self, client_name):
        """ Update manifests for writing signals to device proxies
        Args:
            client_name (str): Name of the client in signal comm
        """
        produces = [{"name": client_name, "signals": [], "signal_groups": []}]
        signal_names = self.signal_names["other"]["outsignals"]
        group_signals = {}
        for signal_spec in self._allowed_specifications(signal_names):
            group = signal_spec[self.dp_position.group.value]
            if group is not None:
                if group not in group_signals:
                    group_signals[group] = []
                group_signals[group].append(
                    self.dp_spec_for_manifest(signal_spec)
                )
            else:
                produces[0]["signals"].append(
                    self.dp_spec_for_manifest(signal_spec)
                )
        for group_name, signals in group_signals.items():
            produces[0]["signal_groups"].append(
                    {"name": group_name,
                     "signals": list(signals)}
            )
        return produces

    @staticmethod
    def cleanup_dp_manifest(manifest):
        """ Remove empty device proxies.
        Args:
            manifest (dict): Device proxy configurations
        """
        if not manifest["produces"][0]["signal_groups"]:
            del manifest["produces"][0]["signal_groups"]
        if not manifest["produces"][0]["signals"]:
            del manifest["produces"][0]["signals"]
        if list(manifest["produces"][0].keys()) == ["name"]:
            del manifest["produces"]
        if not manifest["consumes"][0]["signal_groups"]:
            del manifest["consumes"]
        return manifest



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Module for handling ZoneController composition yaml generation."""

import re
from pathlib import Path
from ruamel.yaml import YAML, scalarstring

from powertrain_build.problem_logger import ProblemLogger
from powertrain_build.types import a2l_range
from powertrain_build.zone_controller.calibration import ZoneControllerCalibration as ZCC


class CompositionYaml(ProblemLogger):
    """Class for handling ZoneController composition yaml generation."""

    def __init__(self, build_cfg, signal_if, unit_cfg, zc_core, zc_dids, zc_nvm, a2l_axis_data, enums):
        """Init.

        Args:
            build_cfg (BuildProjConfig): Object with build configuration settings.
            signal_if (SignalInterfaces): Class holding signal interface information.
            unit_cfg (UnitConfig): Object with unit configurations.
            zc_core (ZCCore): Object with zone controller diagnositic event information.
            zc_dids (ZCDIDs): Object with zone controller diagnostic DID information.
            zc_nvm (ZCNVMDef): Object with NVM definition information.
            a2l_axis_data (dict): Dict with characteristic axis data from A2L file.
            enums (dict): Dict with enum data.
        """
        self.tl_to_autosar_base_types = {
            "Bool": "boolean",
            "Float32": "float32",
            "Int16": "sint16",
            "Int32": "sint32",
            "Int8": "sint8",
            "UInt16": "uint16",
            "UInt32": "uint32",
            "UInt8": "uint8",
        }
        self.build_cfg = build_cfg
        self.unit_src_dirs = build_cfg.get_unit_src_dirs()
        self.composition_spec = signal_if.composition_spec
        self.external_io = signal_if.get_external_io()
        self.unit_cfg = unit_cfg
        self.zc_core = zc_core
        self.zc_dids = zc_dids
        self.zc_nvm = zc_nvm
        self.enums = enums
        self.a2l_axis_data = a2l_axis_data
        base_data_types = self.get_base_data_types()  # Might not be necessary in the long run
        self.data_types = {
            **base_data_types,
            **self.composition_spec.get("data_types", {}),
        }
        self.port_interfaces = self.composition_spec.get("port_interfaces", {})
        self.sharedSwAddrMethod = self.build_cfg.get_composition_config("includeSharedSwAddrMethod")
        if self.sharedSwAddrMethod is not None and not isinstance(self.sharedSwAddrMethod, str):
            self.sharedSwAddrMethod = None
            self.critical("includeSharedSwAddrMethod must be a string if set.")

        calibration_variables, measurable_variables = self._get_variables()
        self.calibration_init_values = self.get_init_values(calibration_variables)
        self.cal_class_info = self._get_class_info(calibration_variables)
        self.meas_class_info = self._get_class_info(measurable_variables)
        self.include_calibration_interface_files = False
        if self.build_cfg.get_code_generation_config(item="generateCalibrationInterfaceFiles"):
            self.include_calibration_interface_files = True
            if self.build_cfg.get_code_generation_config(item="useCalibrationRteMacroExpansion"):
                self.include_calibration_interface_files = False
        if self.include_calibration_interface_files:
            trigger_read_rte_cdata_signal_name = self._get_calibration_trigger_signal_name(calibration_variables)
            self.cal_class_info["tl"]["class_info"].update({
                trigger_read_rte_cdata_signal_name: {
                    "type": ZCC.trigger_read_rte_cdata_signal['data_type'],
                    "width": 1,
                }
            })
            self.cal_class_info["autosar"]["class_info"].update(
                {
                    trigger_read_rte_cdata_signal_name: {
                        "type": ZCC.trigger_read_rte_cdata_signal["data_type"],
                        "access": "READ-WRITE",
                        "init": 0,
                    }
                }
            )

    @staticmethod
    def _cast_init_value(value_str):
        """Cast initialization value to correct type.

        Args:
            value_str (str): String representation of the value.
        Returns:
            (int/float): Value casted to correct type.
        """
        if value_str.endswith("F"):
            return float(value_str[:-1])
        return int(value_str)

    def _prepare_for_xml(self, signal_name, string):
        """Prepare a string for XML serialization.

        Args:
            signal_name (str): The name of the signal.
            string (str): The string to prepare.
        Returns:
            xml_string (str): The prepared string.
        """
        illegal_xml_characters = {
            "&": "&amp;",  # needs to be first in list
            "<": "&lt;",
            ">": "&gt;",
            "'": "&apos;",
            '"': "&quot;"
        }
        xml_string_tmp = "".join(string.splitlines())
        for char, replacement in illegal_xml_characters.items():
            xml_string_tmp = xml_string_tmp.replace(char, replacement)
        if len(xml_string_tmp) > 255:
            self.warning(f"Converted description for {signal_name} exceeds 255 characters and will be truncated.")
            for replacement in illegal_xml_characters.values():
                found = xml_string_tmp.find(replacement, 255-len(replacement), 255+len(replacement))
                if found < 255 and found + len(replacement) > 255:
                    xml_string_tmp = xml_string_tmp[:found]
            xml_string_tmp = xml_string_tmp[:255]  # Since "found" is always < 255 this is safe
        xml_string = scalarstring.DoubleQuotedScalarString(xml_string_tmp)
        return xml_string

    def get_base_data_types(self):
        """Create base data types in expected Autosar/yaml2arxml format."""
        base_data_types = {
            "Bool": {"type": "ENUMERATION", "enums": {"False": 0, "True": 1}},
            "Float32": {
                "type": "FLOAT",
                "limits": {"lower": -3.4e38, "upper": 3.4e38},
            },
        }
        int_data_types = [data_type for data_type in self.tl_to_autosar_base_types if "Int" in data_type]
        for data_type in int_data_types:
            lower, upper = a2l_range(data_type)
            base_data_types[data_type] = {
                "type": "INTEGER",
                "limits": {"lower": lower, "upper": upper},
            }
        return base_data_types

    def generate_yaml(self):
        """Generates a yaml from project/model information."""
        composition_name = self.build_cfg.get_composition_config("compositionName")
        composition_ending = self.build_cfg.get_composition_config("compositionEnding")
        all_info = self.gather_yaml_info()

        output_directory = self.build_cfg.get_src_code_dst_dir()
        self.info(
            "Writing Yaml into %s/%s.%s",
            output_directory,
            composition_name,
            composition_ending,
        )
        Path(output_directory).mkdir(parents=True, exist_ok=True)

        with open(
            f"{output_directory}/{composition_name}.{composition_ending}",
            "w",
            encoding="utf-8",
        ) as file:
            yaml = YAML()
            yaml.width = 1000  # We don't want to wrap lines in the yaml file

            def modify_float_representation(s):
                return re.sub(r"(\s-?\d+)e(?=\+|-\d)", r"\1.e", s)

            yaml.dump(all_info, file, transform=modify_float_representation)

    def gather_yaml_info(self):
        """Creates dict with relevant project/model information.

        Returns:
            all_info (dict): Dict to be written to yaml.
        """
        software_components, pt_build_data_types = self._get_software_components()

        all_info = {
            "ExternalFiles": {
                "Composition": self.build_cfg.get_composition_config("compositionArxml"),
                "GenerateExternalImplementationTypes": self.build_cfg.get_composition_config(
                    "generateExternalImplementationType"
                ),
            },
            "SoftwareComponents": software_components,
            "DataTypes": {**self.data_types, **pt_build_data_types},
            "PortInterfaces": self.port_interfaces,
        }

        return all_info

    def get_init_values(self, calibration_variables):
        """Get initialization values for calibration variables.

        Args:
            calibration_variables (dict): Dict of existing calibration variables.
        Returns:
            init_values (dict): Dictionary with initialization values for calibration variables.
        """
        value_extraction_regexes = [
            (
                re.compile(r"^\s*CVC_CAL[A-Z_]*\s+\w+\s+(?P<name>\w+)\s*=\s*(?P<value>[-\d\.e]+F?)\s*;"),
                lambda regex_match, _: self._cast_init_value(regex_match.group("value")),
            ),
            (
                re.compile(r"^\s*CVC_CAL[A-Z_]*\s+\w+\s+(?P<name>\w+)\[(?P<size>[\d]+)\]\s*=\s*"),
                self._get_array_init_values,
            ),
            (
                re.compile(r"^\s*CVC_CAL[A-Z_]*\s+\w+\s+(?P<name>\w+)\[(?P<rows>[\d]+)\]\[(?P<cols>[\d]+)\]\s*=\s*"),
                self._get_matrix_init_values,
            ),
            (
                re.compile(r"^\s*CVC_CAL[A-Z_]*\s+\w+\s+(?P<name>\w+)\s*=\s*(?P<enum>[a-zA-Z_$][\w_]*?)*;"),
                lambda regex_match, _: regex_match.group("enum"),
            ),
        ]

        init_values = {}
        calibration_definitions = self._get_all_calibration_definitions()
        calibration_definitions.reverse()  # Reverse to pop from the end for performance
        while calibration_definitions:
            line = calibration_definitions.pop()
            for regex, extraction_function in value_extraction_regexes:
                regex_match = regex.match(line)
                if regex_match is not None and regex_match.group("name") in calibration_variables:
                    if regex_match.group("name") in init_values:
                        self.critical("Variable definition for %s already found.", regex_match.group("name"))
                    init_values[regex_match.group("name")] = extraction_function(regex_match, calibration_definitions)

        missing_init_values = set(calibration_variables) - set(init_values.keys())
        if missing_init_values:
            self.critical("Missing init values for calibration variables:\n%s", "\n".join(missing_init_values))

        return init_values

    def _get_all_calibration_definitions(self):
        """Get all calibration definitions from the source files.

        Returns:
            (iter): Iterator with calibration definitions.
        """
        calibration_definitions = []
        end_of_definitions_regex = re.compile(r"^void\s*RESTART_.*")
        c_files = [Path(src_dir, unit.split("__")[0] + ".c").resolve() for unit, src_dir in self.unit_src_dirs.items()]
        for c_file in c_files:
            read_lines = ""
            with c_file.open(mode="r", encoding="latin-1") as file_handle:
                for line in file_handle:
                    if end_of_definitions_regex.match(line):
                        break
                    read_lines += line
            calibration_definitions.extend(re.sub(r"/\*.*?\*/", "", read_lines, flags=re.S).splitlines())
        return calibration_definitions

    def _get_array_init_values(self, array_regex_match, definitions_list):
        """Get initialization values for an array.

        NOTES:
        Modifies the argument definitions_list by popping elements.
        Popping from the end since list is reversed.

        Args:
            array_regex_match (re.Match): Match object with array definition.
            definitions_list (list): List (reversed) with lines to parse.
        Returns:
            (list): List of initialization values for the array.
        """
        array_init_values_str = ""
        line = definitions_list.pop()  # Skip array definition line
        while "};" not in line:
            array_init_values_str += line.strip()
            line = definitions_list.pop()
        array_init_values_str += line.strip()
        array_init_values = re.findall(r"([-\d\.e]+F?),?", array_init_values_str)

        if int(array_regex_match.group("size")) != len(array_init_values):
            self.critical("Could not parse init values for array definition %s.", array_regex_match.group("name"))

        return [self._cast_init_value(value) for value in array_init_values]

    def _get_matrix_init_values(self, matrix_regex_match, definitions_list):
        """Get initialization values for a matrix.

        NOTES:
        Modifies the argument definitions_list by popping elements.
        Popping from the end since list is reversed.

        Args:
            matrix_regex_match (re.Match): Match object with matrix definition.
            definitions_list (list): List (reversed) with lines to parse.
        Returns:
            (list(list)): List of initialization values for the matrix.
        """
        matrix_init_values = []
        matrix_init_values_str = ""
        line = definitions_list.pop()  # Skip matrix definition line
        while "};" not in line:
            matrix_init_values_str += line.strip()
            if "}" in line:
                matrix_init_values.append(re.findall(r"([-\d\.e]+F?),?", matrix_init_values_str))
                matrix_init_values_str = ""
            line = definitions_list.pop()

        row_check = int(matrix_regex_match.group("rows")) != len(matrix_init_values)
        col_check = any(int(matrix_regex_match.group("cols")) != len(row) for row in matrix_init_values)
        if row_check or col_check:
            self.critical("Could not parse init values for matrix definition %s.", matrix_regex_match.group("name"))

        return [[self._cast_init_value(value) for value in row] for row in matrix_init_values]

    def _get_calibration_trigger_signal_name(self, calibration_variables):
        """Get the variable of the calibration trigger.

        Make sure it is not present already.

        Args:
            calibration_variables (dict): Dict of existing calibration variables.
        Returns:
            trigger_signal (str): Name of variable for triggering calibration.
        """
        software_component_name = self.build_cfg.get_composition_config("softwareComponentName")
        trigger_signal = ZCC.trigger_read_rte_cdata_signal["name_template"].format(swc_name=software_component_name)

        if trigger_signal in calibration_variables:
            self.critical("Signal %s already defined in project.", trigger_signal)

        return trigger_signal

    def _edit_event_dict(self, event_dict):
        """Edit event dictionary to use double quoted strings on certain element values.

        The elements that need to be double quoted are JumpDown, JumpUp and EnaDEMInd.
        The values are either "on" or "off", which yaml loader still interprets as boolean.
        Hence the need for double quotes.
        For some reason, "true" is not interpreted as boolean though.

        Args:
            event_dict (dict): Dict with diagnostic event data.
        Returns:
            event_dict (dict): Dict with diagnostic event data,
                updated with double quoted strings on certain element values."""
        for event_data in event_dict.values():
            for key, value in event_data.items():
                if key in ["ACC2", "EnaDEMInd", "JumpDown", "JumpUp"]:
                    event_data[key] = scalarstring.DoubleQuotedScalarString(value)
        return event_dict

    def _get_diagnostic_info(self):
        """Get diagnostic information from composition spec.

        NOTE: This function sets the valid_dids property of the ZCDIDs object.

        Returns:
            diag_dict (dict): Dict containing diagnostic information.
        """
        diag_dict = {}
        diagnostics = self.composition_spec.get("diagnostics", {})

        if self.build_cfg.get_composition_config("includeDiagnostics") == "manual":
            dids = diagnostics.get("dids", {})
            events = self._edit_event_dict(diagnostics.get("events", {}))
        elif self.build_cfg.get_composition_config("includeDiagnostics") == "manual_dids":
            dids = diagnostics.get("dids", {})
            events_tmp = self._edit_event_dict(diagnostics.get("events", {}))
            events = self.zc_core.get_diagnostic_trouble_codes(events_tmp)
        elif self.build_cfg.get_composition_config("includeDiagnostics") == "manual_dtcs":
            self.zc_dids.valid_dids = diagnostics.get("dids", {})
            dids = self.zc_dids.valid_dids
            events = self._edit_event_dict(diagnostics.get("events", {}))
        else:
            self.zc_dids.valid_dids = diagnostics.get("dids", {})
            dids = self.zc_dids.valid_dids
            events_tmp = self._edit_event_dict(diagnostics.get("events", {}))
            events = self.zc_core.get_diagnostic_trouble_codes(events_tmp)
        rids = diagnostics.get("rids", {})

        if dids:
            diag_dict["dids"] = dids
        if events:
            diag_dict["events"] = events
        if rids:
            diag_dict["rids"] = rids
            self.warning("Will not generate code for RIDs, add manually.")
        return diag_dict

    def _get_nvm_info(self):
        """Creates a dict with NVM information.

        NVM dicts also needs to be added to the "static" field in the generated yaml.

        NOTE: This function sets the valid_nvm_definitions property of the ZCNVMDef object.

        Returns:
            nvm_dict (dict): Dict containing NVM information.
            data_types (dict): Dict containing data types for NVM information.
            static_variables (dict): Dict containing "static" variables to add to the "static" field.
        """
        data_types = {}
        static_variables = {}
        yaml_nvm_definitions = self.composition_spec.get("nv-needs", {})
        self.zc_nvm.valid_nvm_definitions = yaml_nvm_definitions
        nvm_dict = self.zc_nvm.valid_nvm_definitions

        for nvm_name, nvm_data in nvm_dict.items():
            init = []
            nr_of_unused_signals = self.zc_nvm.project_nvm_definitions[nvm_name]["size"]
            data_type_name = f"dt_{nvm_name}"
            data_types[data_type_name] = {"type": "RECORD", "elements": {}}
            for signal in self.zc_nvm.project_nvm_definitions[nvm_name]["signals"]:
                element_name = f'{self.zc_nvm.struct_member_prefix}{signal["name"]}'
                nr_of_unused_signals -= signal["x_size"] * signal["y_size"]
                size = max(signal["x_size"], 1) * max(signal["y_size"], 1)
                if size > 1:
                    x_data_type_name = f"dt_{signal['name']}_x"
                    y_data_type_name = f"dt_{signal['name']}_y"
                    if signal["x_size"] > 1 and signal["y_size"] == 1:
                        init.append([0] * signal["x_size"])
                        data_types[data_type_name]["elements"][element_name] = x_data_type_name
                        data_types[x_data_type_name] = {
                            "type": "ARRAY",
                            "size": signal["x_size"],
                            "element": signal["type"],
                        }
                    elif signal["x_size"] > 1 and signal["y_size"] > 1:
                        init.append([[0] * signal["y_size"]] * signal["x_size"])
                        data_types[data_type_name]["elements"][element_name] = y_data_type_name
                        data_types[y_data_type_name] = {
                            "type": "ARRAY",
                            "size": signal["y_size"],
                            "element": x_data_type_name,
                        }
                        data_types[x_data_type_name] = {
                            "type": "ARRAY",
                            "size": signal["x_size"],
                            "element": signal["type"],
                        }
                    else:
                        self.critical("NVM signal size incorrect. x_size should not be 1 if y_size > 1.")
                else:
                    init.append(0)
                    data_types[data_type_name]["elements"][element_name] = signal["type"]

            nvm_data.update({
                "datatype": data_type_name,
                "init": init
            })
            static_variables[nvm_name.lower()] = {
                "type": data_type_name,
                "access": "READ-ONLY",
                "init": init,
            }

            if nr_of_unused_signals > 0:
                # Mimics how we generate the unused member of the structs in nvm_def.py
                nvm_data["init"].append([0] * nr_of_unused_signals)
                data_types[data_type_name]["elements"]["unused"] = f"{data_type_name}_Unused"
                data_types[f"{data_type_name}_Unused"] = {
                    "type": "ARRAY",
                    "size": nr_of_unused_signals,
                    "element": self.zc_nvm.project_nvm_definitions[nvm_name]["default_datatype"],
                }

        return nvm_dict, data_types, static_variables

    def _get_ports_info(self):
        """Creates a dict containing port information.

        Returns:
            ports (dict): Dict containing port information.
        """
        ports = self.composition_spec.get("ports", {})
        for call, call_data in self.composition_spec.get("calls", {}).items():
            if call in ports:
                continue
            ports[call] = {
                "interface": call_data.get("interface", call),
                "direction": call_data["direction"],
            }
        return ports

    def _get_runnable_calls_info(self):
        """Creates a dict containing desired calls for the SWC.

        Returns:
            call_dict(dict): Dict containing runnable calls information.
        """
        call_dict = {}
        for call, call_data in self.composition_spec.get("calls", {}).items():
            call_dict[call] = {"operation": call_data["operation"]}
            if "timeout" in call_data:
                call_dict[call]["timeout"] = call_data["timeout"]
        return call_dict

    def _get_runnable_info(self):
        """Creates a dict containing runnables information.

        Returns:
            dict: Dict containing runnables information.
        """
        autosar_prefix = "AR_"
        swc_prefix = self.build_cfg.get_scheduler_prefix()
        custom_step_function = self.build_cfg.get_composition_config("customYamlStepFunctionName")
        custom_init_function = self.build_cfg.get_composition_config("customYamlInitFunctionName")
        standard_init_function = autosar_prefix + swc_prefix + "VcExtINI"
        init_function = custom_init_function if custom_init_function is not None else standard_init_function
        calibration_variables = list(
            self.cal_class_info["autosar"]["class_info"].keys()
        ) + list(
            self.composition_spec.get("shared", {}).keys()
        )
        swc_content = {
            init_function: {
                "type": "INIT",
                "mode_ref": {
                    "port": "EcuMVccActivationMode",
                    "mode": ["VCC_ACTIVE"],
                    "trigger": "ON-ENTRY"
                },
                "accesses": calibration_variables
            }
        }

        if self.include_calibration_interface_files:
            swc_name = self.build_cfg.get_composition_config("softwareComponentName")
            cal_init_function = autosar_prefix + ZCC.calibration_function_init_template.format(swc_name=swc_name)
            cal_step_function = autosar_prefix + ZCC.calibration_function_step_template.format(swc_name=swc_name)
            swc_content[cal_init_function] = {
                "type": "INIT",
                "mode_ref": {
                    "port": "EcuMVccActivationMode",
                    "mode": ["VCC_ACTIVE"],
                    "trigger": "ON-ENTRY"
                },
                "generateAccessPoints": False,
                "accesses": calibration_variables
            }
            swc_content[cal_step_function] = {
                "type": "PERIODIC",
                "period": 0.1,
                "mode_suppression": [
                    {"port": "EcuMVccActivationMode", "disabled_mode": ["VCC_NOT_ACTIVE"]}
                ],
                "generateAccessPoints": False,
                "accesses": calibration_variables
            }

        call_dict = self._get_runnable_calls_info()
        mode_switch_points_dict = self.composition_spec.get("mode_switch_points", {})
        reads = []
        writes = []
        for port, port_data in self._get_ports_info().items():
            if port_data["direction"] in ["IN", "CLIENT"]:
                reads.append(port)
            else:
                writes.append(port)
        runnables = self.build_cfg.get_units_raster_cfg()["SampleTimes"]

        if len(runnables) == 1 and custom_step_function is not None:
            swc_content[custom_step_function] = {
                "type": "PERIODIC",
                "period": list(runnables.values())[0],
                "mode_suppression": [
                    {"port": "EcuMVccActivationMode", "disabled_mode": ["VCC_NOT_ACTIVE"]}
                ],
                "accesses": calibration_variables,
            }
            if call_dict:
                swc_content[custom_step_function]["calls"] = call_dict
            if mode_switch_points_dict:
                swc_content[custom_step_function]["mode_switch_points"] = mode_switch_points_dict
            if reads:
                swc_content[custom_step_function]["reads"] = reads
            if writes:
                swc_content[custom_step_function]["writes"] = writes
            return swc_content

        if custom_step_function is not None:
            self.warning(
                "Custom step function specified, but multiple runnables defined. Ignoring custom step function."
            )

        for runnable, period in runnables.items():
            key = autosar_prefix + swc_prefix + runnable
            swc_content[key] = {
                "period": period,
                "type": "PERIODIC",
                "mode_suppression": [
                    {"port": "EcuMVccActivationMode", "disabled_mode": ["VCC_NOT_ACTIVE"]}
                ],
                "accesses": calibration_variables,
            }
            if call_dict:
                swc_content[key]["calls"] = call_dict
            if mode_switch_points_dict:
                swc_content[key]["mode_switch_points"] = mode_switch_points_dict
            if reads:
                swc_content[key]["reads"] = reads
            if writes:
                swc_content[key]["writes"] = writes

        return swc_content

    def _get_software_components(self):
        """Creates a dict with swc information and referred data types.

        Returns:
            swcs (dict): SWC information.
            data_types (dict): Data types information.
        """
        software_component_name = self.build_cfg.get_composition_config("softwareComponentName")
        swcs = {software_component_name: {}}
        swcs[software_component_name]["type"] = "SWC"  # Other types than swc??
        swcs[software_component_name]["asil"] = self.build_cfg.get_composition_config("asil")
        swcs[software_component_name]["secure"] = self.build_cfg.get_composition_config("secure")
        swcs[software_component_name]["runnables"] = self._get_runnable_info()
        if self.build_cfg.get_composition_config("includeShared") is True:
            swcs[software_component_name]["shared"] = self.cal_class_info["autosar"]["class_info"]
            for variable_name, variable_info in self.composition_spec.get("shared", {}).items():
                if variable_name in swcs[software_component_name]["shared"]:
                    self.critical("Shared variable %s already defined in project.", variable_name)
                else:
                    swcs[software_component_name]["shared"][variable_name] = variable_info
        elif self.build_cfg.get_composition_config("includeShared") == "manual":
            swcs[software_component_name]["shared"] = {}
            for variable_name, variable_info in self.composition_spec.get("shared", {}).items():
                swcs[software_component_name]["shared"][variable_name] = variable_info
        if self.build_cfg.get_composition_config("includeStatic") is True:
            swcs[software_component_name]["static"] = self.meas_class_info["autosar"]["class_info"]
            for variable_name, variable_info in self.composition_spec.get("static", {}).items():
                if variable_name in swcs[software_component_name]["static"]:
                    self.critical("Static variable %s already defined in project.", variable_name)
                else:
                    swcs[software_component_name]["static"][variable_name] = variable_info
        elif self.build_cfg.get_composition_config("includeStatic") == "manual":
            swcs[software_component_name]["static"] = {}
            for variable_name, variable_info in self.composition_spec.get("static", {}).items():
                swcs[software_component_name]["static"][variable_name] = variable_info
        swcs[software_component_name]["ports"] = self._get_ports_info()
        if self.composition_spec.get("io") is not None:
            swcs[software_component_name]["io"] = self.composition_spec["io"]
        if self.composition_spec.get("ecu") is not None:
            swcs[software_component_name]["ecu"] = self.composition_spec["ecu"]
        diagnostic_info = self._get_diagnostic_info()
        if self.build_cfg.get_composition_config("includeDiagnostics") is not False:
            swcs[software_component_name]["diagnostics"] = diagnostic_info
        nvm_info, nvm_data_types_tmp, static_variables = self._get_nvm_info()
        if self.build_cfg.get_composition_config("includeNvm"):
            swcs[software_component_name]["nv-needs"] = nvm_info
            nvm_data_types = nvm_data_types_tmp
            if self.build_cfg.get_composition_config("includeStatic") is True:
                swcs[software_component_name]["static"].update(static_variables)
        else:
            nvm_data_types = {}

        data_types = {
            **self.cal_class_info["autosar"]["data_types"],
            **self.meas_class_info["autosar"]["data_types"],
            **nvm_data_types,
        }

        return swcs, data_types

    def _get_variables(self):
        """Get calibration and measurable variables from the unit configuration.

        Returns:
            calibration_variables (dict): Dict with calibration variables.
            measurable_variables (dict): Dict with measurable variables.
        """
        calibration_variables = {}
        measurable_variables = {}
        config = self.unit_cfg.get_per_cfg_unit_cfg()
        valid_configs = ["outports", "local_vars", "calib_consts"]
        for valid_config in valid_configs:
            for signal_name, unit_info in config.get(valid_config, {}).items():
                if len(unit_info) > 1:
                    self.critical("Multiple definitions for %s in config json files.", signal_name)
                for info in unit_info.values():
                    if "CVC_CAL" in info["class"]:
                        calibration_variables[signal_name] = info
                    elif "CVC_DISP" in info["class"]:
                        measurable_variables[signal_name] = info
        # External inports should also be considered as measurable variables
        for io_type in self.external_io:
            for signal_name in io_type.get("input", {}).keys():
                for signal_data in config["inports"][signal_name].values():
                    measurable_variables[signal_name] = signal_data
                    continue  # Inports can appear in several units, pick first one
        return calibration_variables, measurable_variables

    def _get_class_info(self, variable_dict):
        """Creates a dict with parameter information and referred data types.

        Args:
            variable_dict (dict): Dictionary with variables and data.
        Returns:
            (dict): Dictionary with variables and data types (Autosar and TL).
        """
        autosar_class_info = {}
        autosar_data_types = {}
        tl_class_info = {}
        for signal_name, info in variable_dict.items():
            (
                autosar_class_info,
                autosar_data_types,
            ) = self._add_autosar_data_types(autosar_class_info, autosar_data_types, signal_name, info)
            if signal_name in autosar_class_info:
                tl_class_info[signal_name] = {
                    "type": info["type"],
                    "autosar_type": autosar_class_info[signal_name]["type"].split("/")[-1],
                    "width": info["width"],
                }
        return {
            "autosar": {
                "class_info": autosar_class_info,
                "data_types": autosar_data_types,
            },
            "tl": {"class_info": tl_class_info, "data_types": {}},
        }

    def _add_autosar_data_types(self, class_info, data_types, signal_name, info):
        """Process a variable for inclusion in composition, adding it's data type to
        data_types and the variable to class_info.

        Args:
            class_info (dict): Dictionary with variables.
            data_types (dict): Dictionary with data types.
            signal_name (string): Name of signal to process.
            info (dict): signal data.
        Returns:

            class_info (dict): Updated dictionary with variables.
            data_types (dict): Updated dictionary with data types.
        """
        if info["type"] in self.enums.keys():
            return class_info, data_types

        isReadOnly = "CVC_DISP" in info["class"] or info["class"] == "CVC_EXT"
        if "Bool" in info["type"]:
            upper = 1
            lower = 0
        else:
            base_type_lower = self.data_types[info["type"]]["limits"]["lower"]
            base_type_upper = self.data_types[info["type"]]["limits"]["upper"]
            lower = info["min"] if info["min"] != "-" else base_type_lower
            upper = info["max"] if info["max"] != "-" else base_type_upper

        if not isinstance(info["width"], list):
            class_info[signal_name] = {
                "type": info["type"],
                "access": "READ-ONLY" if isReadOnly else "READ-WRITE",
                "init": self.calibration_init_values.get(signal_name, max(min(0, upper), lower)),
            }
            if info["description"]:
                class_info[signal_name]["longname"] = self._prepare_for_xml(signal_name, info["description"])
            if info["unit"] and info["unit"] != "-":
                class_info[signal_name]["unit"] = info["unit"]
            if self.sharedSwAddrMethod is not None and not isReadOnly:
                class_info[signal_name]["swAddrMethod"] = f"{self.sharedSwAddrMethod}_{signal_name.split('_')[0]}"
            return class_info, data_types

        if isinstance(lower, list) or isinstance(upper, list):
            if info["width"][0] > 1:
                self.critical(
                    "%s is a multidimentional array of elements with different constraints, not supported.", signal_name
                )
            init = []
            for idx in range(info["width"][1]):
                lower_val = lower[idx] if isinstance(lower, list) else lower
                lower_val = lower_val if lower_val != "-" else base_type_lower
                upper_val = upper[idx] if isinstance(upper, list) else upper
                upper_val = upper_val if upper_val != "-" else base_type_upper
                init.append(max(min(0, upper_val), lower_val))
        else:
            init = max(min(0, upper), lower)
            if info["width"][0] > 1:
                init = [[init] * info["width"][1] for _ in range(info["width"][0])]
            else:
                init = [init] * info["width"][1]

        init = self.calibration_init_values.get(signal_name, init)

        new_data_type = {}
        new_data_type_name = f"dt_{signal_name}"
        if signal_name.startswith("t"):
            if signal_name.endswith("_x"):
                new_data_type_data = {
                    "type": "COM_AXIS",
                    "axis-index": 1,
                    "size": info["width"][1],
                    "limits": {"lower": lower, "upper": upper},
                    "swrecordlayout": {
                        "name": f"Distr_{signal_name}",
                        "type": "INDEX_INCR",
                        "basetype": self.tl_to_autosar_base_types[info["type"]],
                        "label": "X",
                    },
                }
            else:
                axis = self.a2l_axis_data.get(signal_name, {}).get("axes", [signal_name + "_x"])[0]
                new_data_type_data = {
                    "type": "CURVE",
                    "axis": f"dt_{axis}",
                    "limits": {"lower": lower, "upper": upper},
                    "swrecordlayout": {
                        "name": f"Curve_{signal_name}",
                        "type": "COLUMN_DIR",
                        "basetype": self.tl_to_autosar_base_types[info["type"]],
                        "label": "Val",
                    },
                }
            if self.build_cfg.get_composition_config("scaleMapsAndCurves") and "int" in info["type"].lower():
                new_data_type_data["slope"] = info["lsb"]
                new_data_type_data["bias"] = info["offset"]
        elif signal_name.startswith("m"):
            new_data_type_data = {
                "type": "COM_AXIS",
                "size": info["width"][1],
                "limits": {"lower": lower, "upper": upper},
                "swrecordlayout": {
                    "name": f"Distr_{signal_name}",
                    "type": "INDEX_INCR",
                    "basetype": self.tl_to_autosar_base_types[info["type"]],
                },
            }
            if signal_name.endswith("_r"):
                new_data_type_data["axis-index"] = 1
                new_data_type_data["swrecordlayout"]["label"] = "X"
            elif signal_name.endswith("_c"):
                new_data_type_data["axis-index"] = 2
                new_data_type_data["swrecordlayout"]["label"] = "Y"
            else:
                default_names = [signal_name + "_r", signal_name + "_c"]
                axis_r, axis_c = self.a2l_axis_data.get(signal_name, {}).get("axes", default_names)
                new_data_type_data = {
                    "type": "MAP",
                    "x-axis": f"dt_{axis_r}",
                    "y-axis": f"dt_{axis_c}",
                    "limits": {"lower": lower, "upper": upper},
                    "swrecordlayout": {
                        "name": f"Map_{signal_name}",
                        "type": "COLUMN_DIR",
                        "basetype": self.tl_to_autosar_base_types[info["type"]],
                        "label": "Val",
                    },
                }
            if self.build_cfg.get_composition_config("scaleMapsAndCurves") and "int" in info["type"].lower():
                new_data_type_data["slope"] = info["lsb"]
                new_data_type_data["bias"] = info["offset"]
        elif info["width"][0] == 1:
            new_data_type_name = f"dt_{signal_name}_{info['width'][1]}"
            new_data_type_data = {
                "type": "ARRAY",
                "size": info["width"][1],
                "element": info["type"],
            }
        else:
            self.critical("Signal config error for %s.", signal_name)
            return class_info, data_types

        new_data_type[new_data_type_name] = new_data_type_data
        class_info[signal_name] = {
            "type": new_data_type_name,
            "access": "READ-ONLY" if isReadOnly else "READ-WRITE",
            "init": init,
        }
        if info["description"]:
            class_info[signal_name]["longname"] = self._prepare_for_xml(signal_name, info["description"])
        if self.sharedSwAddrMethod is not None and not isReadOnly:
            class_info[signal_name]["swAddrMethod"] = f"{self.sharedSwAddrMethod}_{signal_name.split('_')[0]}"
        data_types = {**data_types, **new_data_type}
        return class_info, data_types



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Module for replacing $CVC_* style references in a2l file."""

import argparse
import re
import sys
from pathlib import Path
from typing import List, Optional


PARSER_HELP = "Replace $CVC_* style references in a2l file"


def configure_parser(parser: argparse.ArgumentParser):
    """Configure parser for CLI."""
    parser.add_argument("a2l_target_file")
    parser.set_defaults(func=replace_tab_verb_cli)


def replace_tab_verb_cli(args: argparse.Namespace):
    """CLI wrapper function for passing in Namespace object.

    This allows maintaining a standardized CLI function signature while not breaking backwards
    compatibility with replace_tab_verb.
    """
    replace_tab_verb(args.a2l_target_file)


def replace_tab_verb(file_path: Path):
    """Replace custom Units with conversion table."""
    with open(file_path, encoding='ISO-8859-1') as f_h:
        a2l_text = f_h.read()

    ##############################################################
    # Absolutely horrible multi-line regex                       #
    #                                                            #
    # Translates a fake COMPU_METHOD                             #
    # with custom $CVC_* Unit into using a conversion table      #
    #                                                            #
    ##############################################################
    # If a dollar is found make the following changes:           #
    # 'RAT_FUNC' => 'TAB_VERB'                                   #
    # 'LINEAR' => 'TAB_VERB'                                     #
    # New format => "%12.6"                                      #
    # Save the $-tag and remove it from the unit                 #
    # Change the line below the '/* Unit */'- line to            #
    # "['COMPU_TAB_REF CONV_TAB_' $-tag]                         #
    ##############################################################
    # ***************************BEFORE************************* #
    ##############################################################
    # /begin COMPU_METHOD                                        #
    #     VcDummy_spm_1_0_0_None__CVC_EmCoBaseMode   /* Name */  #
    #     ""    /* LongIdentifier */                             #
    #     RAT_FUNC    /* ConversionType */                       #
    #     "%11.3"   /* Format */                                 #
    #     "-,$CVC_EmCoBaseMode" /* Unit */                       #
    #     COEFFS 0 1 0.0 0 0 1                                   #
    # /end COMPU_METHOD                                          #
    ##############################################################
    # ***************************AFTER************************** #
    ##############################################################
    # /begin COMPU_METHOD                                        #
    #     VcDummy_spm_1_0_0_None__CVC_EmCoBaseMode   /* Name */  #
    #     ""    /* LongIdentifier */                             #
    #     TAB_VERB    /* ConversionType */                       #
    #     "%12.6"   /* Format */                                 #
    #     "-" /* Unit */                                         #
    #     COMPU_TAB_REF CONV_TAB_CVC_EmCoBaseMode               #
    # /end COMPU_METHOD                                          #
    ##############################################################

    COMPU_METHOD_EXPRESSION = (
        r"(RAT_FUNC|LINEAR)"  # Match conversion types RAT_FUNC or LINEAR
        r"(?P<conv_type_end>\s+/\* ConversionType \*/\n)"  # Capture end of conversion type line
        r'(?P<format_begin>\s+"%)'  # Capture beginning of format line
        r"[0-9]+[.][0-9]+"  # Capture formatting
        r'(?P<format_end>"\s+/\* Format \*/\n)'  # Capture end format line
        r'(?P<unit_begin>\s+"[\sA-Za-z0-9\-_/]+)'  # Capture beginning of unit line
        r",\$(?P<tab_verb>CVC_[A-Za-z0-9_]+)"  # Capture tab_verb 'CVC_*' of $CVC_EmCoBaseMode
        r'(?P<unit_end>"\s*/\* Unit \*/\n)'  # Capture end of unit line
        r"\s+COEFFS( \d+(.\d+)?)+"
    )  # Capture COEFFS line
    COMPU_METHOD_REGEX = re.compile(COMPU_METHOD_EXPRESSION)
    TAB_VERB_REPLACE = (
        r"TAB_VERB\g<conv_type_end>\g<format_begin>12.6\g<format_end>\g<unit_begin>\g<unit_end>"
        r"        COMPU_TAB_REF CONV_TAB_\g<tab_verb>"
    )

    a2l_patched = COMPU_METHOD_REGEX.sub(TAB_VERB_REPLACE, a2l_text)
    with open(file_path, "w") as f_h:
        f_h.write(a2l_patched)


def main(argv: Optional[List[str]] = None):
    """Main function for CLI."""
    parser = argparse.ArgumentParser(description=PARSER_HELP)
    configure_parser(parser)
    args = parser.parse_args(argv)
    args.func(args)


if __name__ == "__main__":
    main(sys.argv[1:])



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

# -*- coding: utf-8 -*-

"""Python module for abstracting Pybuild applications"""
from pathlib import Path
import json

from powertrain_build.interface.base import BaseApplication, Signal, MultipleProducersError, Domain, Interface
from powertrain_build.lib import logger
from powertrain_build.build_proj_config import BuildProjConfig
from powertrain_build.feature_configs import FeatureConfigs
from powertrain_build.unit_configs import UnitConfigs
from powertrain_build.user_defined_types import UserDefinedTypes


LOGGER = logger.create_logger("application")


def get_raster_to_raster_interfaces(rasters):
    """Generate a list of Interfaces for internal raster-to-raster signals.

    Args:
        rasters (list): Input rasters (from app.get_rasters())
    Returns:
        interfaces (list(interfaces)): List of unique raster-to-raster-interfaces.
    """
    raster_pairs = []
    for current_raster in rasters:
        for corresponding_raster in [r for r in rasters if r != current_raster]:
            # If we have interface a_b, no need to produce b_a.
            if (corresponding_raster, current_raster) not in raster_pairs:
                raster_pairs.append((current_raster, corresponding_raster))

    return [Interface(raster[0], raster[1]) for raster in raster_pairs]


def get_internal_domain(rasters):
    """ Create an internal domain of signals

    Loops through all raster<->raster communications and adds them to a domain object

    Args:
        rasters (list(Raster)): rasters to calculate communication for
    Returns:
        domain (Domain): signals belonging to the same domain
    """
    internal = Domain()
    internal.set_name("internal")
    for interface in get_raster_to_raster_interfaces(rasters):
        internal.add_interface(interface)

    return internal


def get_active_signals(signals, feature_cfg):
    """ Filter out inactive signals. """
    LOGGER.debug('Filtering %s', signals)
    return [signal for signal in signals if feature_cfg.check_if_active_in_config(signal.properties['configs'])]


class Application(BaseApplication):
    """ Object for holding information about a pybuild project """
    def __init__(self):
        self.name = None
        self.pybuild = {'build_cfg': None,
                        'feature_cfg': None,
                        'unit_vars': {}}
        self._insignals = None
        self._outsignals = None
        self._signals = None
        self._raster_definitions = []
        self._services = None
        self._methods = []
        self._enumerations = None
        self._structs = None

    def parse_definition(self, definition):
        """ Parse ProjectCfg.json, get code switch values and read config.json files.
        Add the information to the object.

        Args:
            definition (Path): Path to ProjectCfg.json
        """
        self.pybuild['build_cfg'] = BuildProjConfig(str(definition))
        self.name = self.pybuild['build_cfg'].name
        self.pybuild['feature_cfg'] = FeatureConfigs(self.pybuild['build_cfg'])
        unit_cfg = UnitConfigs(self.pybuild['build_cfg'], self.pybuild['feature_cfg'])
        self.pybuild['unit_vars'] = unit_cfg.get_per_unit_cfg()
        self.pybuild['user_defined_types'] = UserDefinedTypes(self.pybuild['build_cfg'], unit_cfg)

    def get_domain_names(self):
        """ Get domain names. """
        return self.pybuild['build_cfg'].device_domains.values()

    def get_domain_mapping(self):
        """ Get device to signal domain mapping. """
        return self.pybuild['build_cfg'].device_domains

    def get_methods(self):
        """ Get csp methods. """
        if self._signals is None:
            self._get_signals()
        return self._methods

    @property
    def enumerations(self):
        """ Get enumerations defined in the project. """
        if self._enumerations is None:
            self._enumerations = self.pybuild['user_defined_types'].get_enumerations()
        return self._enumerations

    @property
    def structs(self):
        """ Get structs defined in the project. """
        if self._structs is None:
            self._structs = self.pybuild['user_defined_types'].get_structs()
        return self._structs

    @property
    def services(self):
        """ Get interface to service mapping. """
        if self._services is None:
            services_file = self.get_services_file()
            self._services = self.pybuild['build_cfg'].get_services(services_file)
        return self._services

    def get_service_mapping(self):
        """ Get interface to service mapping. """
        return self.services

    def get_services_file(self):
        """ Get path to file specifying interface to service mapping. """
        return self.pybuild['build_cfg'].services_file

    def get_name(self, definition):
        """ Parse ProjectCfg.json and return the specified project name """
        if self.name is None:
            return BuildProjConfig(str(definition)).name
        return self.name

    def _get_signals(self):
        """ Calculate parse all inport and outports of all models """
        self._insignals = set()
        self._outsignals = set()
        defined_ports = {'inports': set(), 'outports': set()}
        for unit, data in self.pybuild['unit_vars'].items():
            self.parse_ports(data, defined_ports, self.pybuild['feature_cfg'], unit)
            self.parse_csp_methods(data, self.pybuild['feature_cfg'], unit)

    def parse_ports(self, port_data, defined_ports, feature_cfg, unit):
        """ Parse ports for one model, based on code switch values.
        Modifies the defined_ports dict and the object.

        Args:
            port_data (dict): port data for a model/unit
            defined_ports (set): all known signals
            feature_cfg (FeatureConfigs): pybuild parsed object for code switches
            unit (string): Name of model/unit
        """
        if self._signals is None:
            self._signals = {}
        for port_type, outport in {'outports': True, 'inports': False}.items():
            for port_name, data in port_data.get(port_type, {}).items():
                # Get what signals we are dealing with
                if not feature_cfg.check_if_active_in_config(data['configs']):
                    continue
                if port_name not in self._signals:
                    signal = Signal(port_name, self)
                    self._signals.update({port_name: signal})
                else:
                    signal = self._signals[port_name]
                # Add information about which models are involved while we are reading it
                if outport:
                    try:
                        signal.set_producer(unit)
                    except MultipleProducersError as mpe:
                        LOGGER.debug(mpe.message)
                        signal.force_producer(unit)
                    self._outsignals.add(port_name)
                else:
                    signal.consumers = unit
                    self._insignals.add(port_name)
                defined_ports[port_type].add(port_name)

    def parse_csp_methods(self, port_data, feature_cfg, unit):
        """ Parse csp methods.

        Args:
            port_data (dict): port data for a model/unit.
            feature_cfg (FeatureConfigs): pybuild parsed object for code switches
            unit (string): Name of model/unit
        """
        if self._signals is None:
            self._signals = {}
        methods = port_data.get('csp', {}).get('methods', {})
        for method_name, data in methods.items():
            if feature_cfg.check_if_active_in_config(data['configs']):
                method = Method(self, unit)
                method.parse_definition((method_name, data))
                self._methods.append(method)

    def get_signal_properties(self, signal):
        """ Get properties for the signal from powertrain_build definition.

        Args:
            signal (Signal): Signal object
        Returns:
            properties (dict): Properties of the signal in pybuild
        """
        # Hack: Take the first consumer or producer if any exists
        for producer in signal.producer:
            return self.pybuild['unit_vars'][producer]['outports'][signal.name]
        for consumer in signal.consumers:
            return self.pybuild['unit_vars'][consumer]['inports'][signal.name]
        return {}

    def get_rasters(self):
        """ Get rasters parsed from powertrain_build.

        Returns:
            rasters (list): rasters parsed from powertrain_build
        """
        if self._signals is None:
            self._get_signals()
        raster_definition = self.pybuild['build_cfg'].get_units_raster_cfg()
        rasters = []
        for raster_field, raster_content in raster_definition.items():
            if raster_field in ['SampleTimes']:
                continue
            for name, content in raster_content.items():
                if name in ['NoSched']:
                    continue
                raster = Raster(self)
                raster.parse_definition((name, content, self._signals))
                rasters.append(raster)
        return rasters

    def get_models(self):
        """ Get models and parse their config files.

        Returns:
            models (list(Model)): config.jsons parsed
        """
        rasters = self.get_rasters()
        # Since one model can exist in many rasters. Find all unique model names first.
        cfg_dirs = self.pybuild['build_cfg'].get_unit_cfg_dirs()
        model_names = set()
        for raster in rasters:
            model_names = model_names.union(raster.models)
        models = []
        for model_name in model_names:
            if model_name not in cfg_dirs:
                LOGGER.debug("%s is generated code. It does not have a config.", model_name)
                continue
            model = Model(self)
            cfg_dir = cfg_dirs[model_name]
            config = Path(cfg_dir, f'config_{model_name}.json')
            model.parse_definition((model_name, config))
            models.append(model)
        return models

    def get_translation_files(self):
        """ Find all yaml files in translation file dirs.

        Returns:
            translation_files (list(Path)): translation files
        """
        translation_files = []
        cfg_dirs = self.pybuild['build_cfg'].get_translation_files_dirs()
        for cfg_dir in cfg_dirs.values():
            cfg_path = Path(cfg_dir)
            translation_files.extend(cfg_path.glob('*.yaml'))
        translation_files = list(set(translation_files))
        return translation_files


class Raster(BaseApplication):
    """ Object for holding information about a raster """
    def __init__(self, app):
        """Construct a new Raster object.

        Args:
            app (powertrain_build.interface.application.Application): Pybuild project raster is part of
        """
        self.app = app
        self.name = str()
        self._insignals = None
        self._outsignals = None
        self._available_signals = None
        self.models = set()

    def parse_definition(self, definition):
        """ Parse the definition from powertrain_build.

        Args:
            definition (tuple):
                name (string): Name of the raster
                content (list): Models in the raster
                app_signals (dict): All signals in all rasters
        """
        self.name = definition[0]
        self.models = set(definition[1])
        self._available_signals = definition[2]

    def _get_signals(self):
        """ Add signals from the project to the raster if they are used here

        Modifies the object itself.
        """
        self._insignals = set()
        self._outsignals = set()
        self._signals = {}
        if self._available_signals is None:
            return
        for signal in self._available_signals.values():
            for consumer in signal.consumers:
                if consumer in self.models:
                    self._signals.update({signal.name: signal})
                    self._insignals.add(signal.name)
            if isinstance(signal.producer, set):
                for producer in signal.producer:
                    if producer in self.models:
                        self._signals.update({signal.name: signal})
                        self._outsignals.add(signal.name)
            else:
                if signal.producer in self.models:
                    self._signals.update({signal.name: signal})
                    self._outsignals.add(signal.name)

    def get_signal_properties(self, signal):
        """ Get properties for the signal from powertrain_build definition.

        Args:
            signal (Signal): Signal object
        Returns:
            properties (dict): Properties of the signal in pybuild
        """
        for producer in signal.producer:
            if producer in self.app.pybuild['unit_vars']:
                return self.app.get_signal_properties(signal)
        return {}


class Model(BaseApplication):
    """ Object for holding information about a model """
    def __init__(self, app):
        self.app = app
        self.name = str()
        self.config = None
        self._insignals = None
        self._outsignals = None
        self._signal_specs = None

    def get_signal_properties(self, signal):
        """ Get properties for the signal from powertrain_build definition.

        Args:
            signal (Signal): Signal object
        Returns:
            properties (dict): Properties of the signal in pybuild
        """
        if self._signal_specs is None:
            self._get_signals()
        if signal.name in self._signal_specs:
            return self._signal_specs[signal.name]
        return {}

    def _get_signals(self):
        """ Add signals from the project to the model if they are used here

        Modifies the object itself.
        Entrypoint for finding signals from the base class.
        """
        self._insignals = set()
        self._outsignals = set()
        self._signals = {}
        self._signal_specs = {}
        self._parse_unit_config(self.config)

    def _parse_unit_config(self, path):
        """ Parse a unit config file.

        Broken out of get_signals to be recursive for included configs.
        """
        cfg = self._load_json(path)
        for signal_spec in cfg['inports'].values():
            signal = Signal(signal_spec['name'], self)
            self._insignals.add(signal.name)
            self._signals.update({signal.name: signal})
            self._signal_specs[signal.name] = signal_spec
        for signal_spec in cfg['outports'].values():
            signal = Signal(signal_spec['name'], self)
            self._outsignals.add(signal.name)
            self._signals.update({signal.name: signal})
            self._signal_specs[signal.name] = signal_spec
        for include_cfg in cfg.get('includes', []):
            LOGGER.debug('%s includes %s in %s', self.name, include_cfg, path.parent)
            include_path = Path(path.parent, f'config_{include_cfg}.json')
            self._parse_unit_config(include_path)

    @staticmethod
    def _load_json(path):
        """ Small function that opens and loads a json file.

        Exists to be mocked in unittests
        """
        with open(path, encoding="utf-8") as fhndl:
            return json.load(fhndl)

    def parse_definition(self, definition):
        """ Parse the definition from powertrain_build.

        Args:
            definition (tuple):
                name (string): Name of the model
                configuration (Path): Path to config file
        """
        self.name = definition[0]
        self.config = definition[1]
        self._get_signals()


class Method(BaseApplication):
    """ Object for holding information about a csp method call """
    def __init__(self, app, unit):
        """Construct a new Method object.

        Args:
            app (powertrain_build.interface.application.Application): Pybuild project raster is part of.
            unit (str): Model that the method is defined in.
        """
        self.app = app
        self.unit = unit
        self.name = str()
        self.namespace = str()
        self.adapter = str()
        self.description = None
        self._signals = {}
        self._insignals = set()
        self._outsignals = set()
        self._primitives = {}
        self._properties = {}

    def parse_definition(self, definition):
        """ Parse the definition from powertrain_build.

        Args:
            definition (tuple):
                name (string): Name of the model
                configuration (dict): Configuration of method
        """
        name = definition[0]
        configuration = definition[1]
        self.name = name
        self.adapter = configuration['adapter']
        self.namespace = configuration['namespace']
        self._primitives[name] = configuration['primitive']
        if 'description' in configuration:
            self.description = configuration['description']
        signals = configuration.get('ports', {})
        outsignals = signals.get('out', {})
        for signal_name, signal_data in outsignals.items():
            signal = self._add_signal(signal_name)
            signal.consumers = name
            signal.set_producer(name)
            self._primitives[signal_name] = signal_data['primitive']
            self._properties[signal_name] = signal_data
            self._outsignals.add(signal_name)
        insignals = signals.get('in', {})
        for signal_name, signal_data in insignals.items():
            signal = self._add_signal(signal_name)
            signal.consumers = name
            signal.set_producer(name)
            self._insignals.add(signal_name)
            self._primitives[signal_name] = signal_data['primitive']
            self._properties[signal_name] = signal_data

    def _add_signal(self, signal_name):
        """ Add a signal used by the method.

        Args:
            signal_name (str): Name of the signal
        """
        if signal_name not in self._signals:
            signal = Signal(signal_name, self)
            self._signals.update({signal_name: signal})
        else:
            signal = self._signals[signal_name]
        return signal

    def get_signal_properties(self, signal):
        """ Get properties for the signal from csp method configuration.

        Args:
            signal (Signal): Signal object
        Returns:
            properties (dict): Properties of the signal in pybuild
        """
        return self._properties[signal.name]

    def get_primitive(self, primitive_name):
        """ Get primitive.

        Args:
            primitive_name (str): Name of primitive part
        Returns:
            primitive (str): Primitive
        """
        return self._primitives[primitive_name]



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

# -*- coding: utf-8 -*-
"""Module used to read project and base configuration files and provides methods for abstraction."""

import glob
import json
import os
import shutil
import pathlib
from pprint import pformat

from powertrain_build.lib.helper_functions import deep_dict_update
from powertrain_build.versioncheck import Version


class BuildProjConfig:
    """A class holding build project configurations."""

    def __init__(self, prj_config_file):
        """Read project configuration file to internal an representation.

        Args:
            prj_config_file (str): Project config filename
        """
        super().__init__()
        self._prj_cfg_file = prj_config_file
        prj_root_dir, _ = os.path.split(prj_config_file)
        self._prj_root_dir = os.path.abspath(prj_root_dir)

        with open(prj_config_file, 'r', encoding="utf-8") as pcfg:
            self._prj_cfg = json.load(pcfg)
        if not Version.is_compatible(self._prj_cfg.get('ConfigFileVersion')):
            raise ValueError('Incompatible project config file version.')
        # Load a basic config that can be common for several projects
        # the local config overrides the base config
        if 'BaseConfig' in self._prj_cfg:
            fil_tmp = os.path.join(prj_root_dir, self._prj_cfg['BaseConfig'])
            fil_ = os.path.abspath(fil_tmp)
            with open(fil_, 'r', encoding="utf-8") as bcfg:
                base_cnfg = json.load(bcfg)
            deep_dict_update(self._prj_cfg, base_cnfg)
            if not Version.is_compatible(self._prj_cfg.get('BaseConfigFileVersion')):
                raise ValueError('Incompatible base config file version.')
        deep_dict_update(self._prj_cfg, self._get_code_generation_config())
        self._composition_config = self._parse_composition_config(self._prj_cfg.get('CompositionConfig', {}))
        self.has_yaml_interface = self._prj_cfg['ProjectInfo'].get('yamlInterface', False)
        self.device_domains = self._get_device_domains()
        self.services_file = self._get_services_file()
        self._load_unit_configs()
        self._add_global_const_file()
        self._all_units = []
        self._calc_all_units()
        self.name = self._prj_cfg['ProjectInfo']['projConfig']
        self.allow_undefined_unused = self._prj_cfg['ProjectInfo'].get('allowUndefinedUnused', True)
        self._scheduler_prefix = self._prj_cfg['ProjectInfo'].get('schedulerPrefix', '')
        if self._scheduler_prefix:
            self._scheduler_prefix = self._scheduler_prefix + '_'

    def __repr__(self):
        """Get string representation of object."""
        return pformat(self._prj_cfg['ProjectInfo'])

    def _get_default_code_generation_config(self):
        return {
            'generalAsilLevelDebug': 'B',
            'generalAsilLevelDependability': 'B',
            'generateCalibrationInterfaceFiles': False,
            'useCalibrationRteMacroExpansion': False,
            'generateCoreDummy': False,
            'generateDummyVar': False,
            'generateInterfaceHeaders': False,
            'rteCheckpointIdSize': None,
            'customRteCheckpointEntityName': 'AR_{raster}_CheckpointReached',
            'generateYamlInterfaceFile': False,
            'includeAllEnums': False,
            'mapToRteEnums': False,
            'propagateTagName': False,
            'useA2lSymbolLinks': False,
            'useRteNvmStructs': False,
            'useCamelCaseForNvmVariables': False
        }

    def _get_code_generation_config(self):
        """ Get code generation configuration.

        Anything already set in CodeGenerationConfig in ProjectCfg.json takes priority.
        If there is a project template for the ECU supplier, those values are used,
        unless already set in ProjectCfg.json.
        Finally, default values are inserted for missing keys.

        Args:
            item (str): Item to get from the configuration. If None, the whole configuration is returned.
        Returns:
            (dict): Code generation configuration.
        """
        code_generation_configuration = {}
        ecu_supplier = self.get_ecu_info()[0]
        deep_dict_update(code_generation_configuration, self._prj_cfg.get('ProjectTemplates', {}).get(ecu_supplier, {}))
        deep_dict_update(code_generation_configuration, self._get_default_code_generation_config())
        return {'CodeGenerationConfig': code_generation_configuration}

    def _get_device_domains(self):
        file_name = self._prj_cfg['ProjectInfo'].get('deviceDomains')
        full_path = pathlib.Path(self._prj_root_dir, file_name)
        if full_path.is_file():
            with open(full_path, 'r', encoding="utf-8") as device_domains:
                return json.loads(device_domains.read())
        return {}

    def _get_services_file(self):
        file_name = self._prj_cfg['ProjectInfo'].get('serviceInterfaces', '')
        full_path = pathlib.Path(self._prj_root_dir, file_name)
        return full_path

    @staticmethod
    def get_services(services_file):
        """Get the services from the services file.

        Args:
            services_file (pathlib.Path): The services file.

        Returns:
            (dict): The services.
        """
        if services_file.is_file():
            with services_file.open() as services:
                return json.loads(services.read())
        return {}

    def _load_unit_configs(self):
        """Load Unit config json file.

        This file contains which units are included in which projects.
        """
        if 'UnitCfgs' in self._prj_cfg:
            fil_tmp = os.path.join(self._prj_root_dir, self._prj_cfg['UnitCfgs'])
            with open(fil_tmp, 'r', encoding="utf-8") as fpr:
                tmp_unit_cfg = json.load(fpr)
                sample_times = tmp_unit_cfg.pop('SampleTimes')
                self._unit_cfg = {
                    'Rasters': tmp_unit_cfg,
                    'SampleTimes': sample_times
                }
        else:
            raise ValueError('UnitCfgs is not specified in project config')

    def _add_global_const_file(self):
        """Add the global constants definition to the 'not_scheduled' time raster."""
        ugc = self.get_use_global_const()
        if ugc:
            self._unit_cfg['Rasters'].setdefault('NoSched', []).append(ugc)

    def create_build_dirs(self):
        """Create the necessary output build dirs if they are missing.

        Clear the output build dirs if they exist.
        """
        src_outp = self.get_src_code_dst_dir()
        if os.path.exists(src_outp):
            shutil.rmtree(src_outp)
        os.makedirs(src_outp)

        log_outp = self.get_log_dst_dir()
        if os.path.exists(log_outp):
            shutil.rmtree(log_outp)
        os.makedirs(log_outp)

        rep_outp = self.get_reports_dst_dir()
        if os.path.exists(rep_outp):
            shutil.rmtree(rep_outp)
        os.makedirs(rep_outp)

        unit_cfg_outp = self.get_unit_cfg_deliv_dir()
        if os.path.exists(unit_cfg_outp):
            shutil.rmtree(unit_cfg_outp)
        if unit_cfg_outp is not None:
            os.makedirs(unit_cfg_outp)

    def get_code_generation_config(self, item=None):
        """ Get code generation configuration.

        Args:
            item (str): Item to get from the configuration. If None, the whole configuration is returned.
        Returns:
            (dict): Code generation configuration.
        """
        if item is not None:
            return self._prj_cfg['CodeGenerationConfig'].get(item, {})
        return self._prj_cfg['CodeGenerationConfig']

    def get_memory_map_config(self):
        """ Get memory map configuration.

        Returns:
            (dict): Memory map configuration.
        """
        return self._prj_cfg.get('MemoryMapConfig', {})

    def get_a2l_cfg(self):
        """ Get A2L configuration from A2lConfig.

        Returns:
            config (dict): A2L configuration
        """
        a2l_config = self._prj_cfg.get('A2lConfig', {})
        return {
            'name': a2l_config.get('name', self._prj_cfg["ProjectInfo"]["projConfig"]),
            'allow_kp_blob': a2l_config.get('allowKpBlob', True),
            'ip_address':  a2l_config.get('ipAddress', "169.254.4.10"),
            'ip_port': '0x%X' % a2l_config.get('ipPort', 30000),
            'asap2_version': a2l_config.get('asap2Version', "1 51")
        }

    def get_enable_end_to_end_status_signals(self):
        """Get the enable end-to-end status signals configuration.

        NOTE: Only appicable for device proxy type signal interfaces.

        Returns:
            (bool): True if end-to-end status signals are enabled, False otherwise.
        """
        return self._prj_cfg['ProjectInfo'].get('enableEndToEndStatusSignals', False)

    def get_unit_cfg_deliv_dir(self):
        """Get the directory where to put the unit configuration files.

        If this key is undefined, or set to None, the unit-configs will
        not be copied to the output folder.

        Returns:
            A path to the unit deliver dir, or None

        """
        if 'unitCfgDeliveryDir' in self._prj_cfg['ProjectInfo']:
            return os.path.join(self.get_root_dir(),
                                os.path.normpath(self._prj_cfg['ProjectInfo']
                                                 ['unitCfgDeliveryDir']))
        return None

    def get_root_dir(self):
        """Get the root directory of the project.

        Returns:
            A path to the project root (with wildcards)

        """
        return self._prj_root_dir

    def get_src_code_dst_dir(self):
        """Return the absolute path to the source output folder."""
        return os.path.join(self.get_root_dir(),
                            os.path.normpath(self._prj_cfg['ProjectInfo']
                                             ['srcCodeDstDir']))

    def _parse_composition_config(self, file_config):
        """Parse the composition configuration from project config."""
        composition_config = {
            'compositionName': None,
            'compositionEnding': 'yml',
            'compositionArxml': file_config.get("compositionArxml", None),
            'customYamlInitFunctionName': file_config.get("customYamlInitFunctionName", None),
            'customYamlStepFunctionName': file_config.get("customYamlStepFunctionName", None),
            'generateExternalImplementationType': file_config.get("generateExternalImplementationType", True),
            'softwareComponentName': file_config.get("softwareComponentName", self.get_a2l_cfg()['name']),
            'asil': file_config.get('asil', 'QM'),
            'secure': file_config.get('secure', False),
            'includeStatic': file_config.get('includeStatic', True),
            'includeShared': file_config.get('includeShared', True),
            'includeSharedSwAddrMethod': file_config.get('includeSharedSwAddrMethod', None),
            'includeDiagnostics': file_config.get('includeDiagnostics', True),
            'includeNvm': file_config.get('includeNvm', True),
            'scaleMapsAndCurves': file_config.get('scaleMapsAndCurves', True),
        }
        composition_name = file_config.get("compositionName", None)
        if composition_name is not None:
            composition_config["compositionName"] = composition_name.split(".")[0]
            if "." in composition_name:
                composition_config["compositionEnding"] = composition_name.split(".")[1]
        return composition_config

    def get_composition_config(self, key=None):
        """Get the composition configuration from project config."""
        if key is None:
            return self._composition_config
        return self._composition_config[key]

    def get_car_com_dst(self):
        """Return the absolute path to the source output folder."""
        return os.path.join(self.get_root_dir(),
                            os.path.normpath(self._prj_cfg['ProjectInfo']
                                             ['didCarCom']))

    def get_reports_dst_dir(self):
        """Get the destination dir for build reports.

        Returns:
            A path to the report files destination directory (with wildcards)

        """
        return os.path.join(self.get_root_dir(),
                            os.path.normpath(self._prj_cfg['ProjectInfo']
                                             ['reportDstDir']))

    def get_all_reports_dst_dir(self):
        """Get the destination dir for build reports.

        Returns:
            A path to the report files destination directory (for all projects)

        """
        return os.path.join(self.get_root_dir(), "..", "..", "Reports")

    def get_log_dst_dir(self):
        """Return the absolute path to the log output folder.

        Returns:
            A path to the log files destination directory (with wildcards)

        """
        return os.path.join(self.get_root_dir(),
                            os.path.normpath(self._prj_cfg['ProjectInfo']
                                             ['logDstDir']))

    def get_core_dummy_name(self):
        """Return the file name of the core dummy file from the config file.

        Returns:
            A file name for the core dummy files

        """
        path = os.path.join(self.get_src_code_dst_dir(),
                            os.path.normpath(self._prj_cfg['ProjectInfo']
                                             ['coreDummyFileName']))
        return path

    def get_feature_conf_header_name(self):
        """Return the feature configuration header file name.

        Returns:
            A file name for the feature config header file

        """
        return self._prj_cfg['ProjectInfo']['featureHeaderName']

    def get_ts_header_name(self):
        """Return the name of the ts header file, defined in the config file.

        Returns:
            The file name of the file defining all unit raster times

        """
        return self._prj_cfg['ProjectInfo']['tsHeaderName']

    def get_included_units(self):
        """Return a list of all the included units in the project.

        TODO:Consider moving this to the Feature Configs class if we start
        using our configuration tool for model inclusion and scheduling
        TODO:Consider calculate this on init and storing the result in the
        class. this method would the just return the stored list.
        """
        units_dict = self._unit_cfg['Rasters']
        units = []
        for unit in units_dict.values():
            units.extend(unit)
        return units

    def get_included_common_files(self):
        """Return a list of all the included common files in the project.

        Returns:
            included_common_files ([str]): The names of the common files which are included in the project.

        """
        return self._prj_cfg.get('includedCommonFiles', [])

    def _calc_all_units(self):
        """Return a list of all the units."""
        units = set()
        for runits in self._unit_cfg['Rasters'].values():
            units = units.union(set(runits))
        self._all_units = list(units)

    def get_includes_paths_flat(self):
        """Return list of paths to files to be included flat in source directory."""
        includes_paths = self._prj_cfg.get('includesPaths', [])
        return [os.path.join(self.get_root_dir(), os.path.normpath(path)) for path in includes_paths]

    def get_includes_paths_tree(self):
        """Return list of paths to files to included with directories in source directory."""
        includes_paths_tree = self._prj_cfg.get('includesPathsTree', [])
        return [os.path.join(self.get_root_dir(), os.path.normpath(path)) for path in includes_paths_tree]

    def get_all_units(self):
        """Return a list of all the units."""
        return self._all_units

    def get_prj_cfg_dir(self):
        """Return the directory containing the project configuration files.

        Returns:
            An absolute path to the project configuration files

        """
        return os.path.join(self._prj_root_dir,
                            self._prj_cfg['ProjectInfo']['configDir'])

    def get_scheduler_prefix(self):
        """Returns a prefix used to distinguish function calls in one project from
        similarly named functions in other projects, when linked/compiled together

        Returns:
        scheduler_prefix (string): prefix for scheduler functions.
        """
        return self._scheduler_prefix

    def get_local_defs_name(self):
        """Return a string which defines the file name of local defines.

        Returns:
            A string containing the wildcard file name local defines

        """
        return self._prj_cfg['ProjectInfo']['prjLocalDefs']

    def get_codeswitches_name(self):
        """Return a string which defines the file name of code switches.

        Returns:
            A string containing the wildcard file name code switches

        """
        return self._prj_cfg['ProjectInfo']['prjCodeswitches']

    def get_did_cfg_file_name(self):
        """Return the did definition file name.

        Returns:
            DID definition file name

        """
        return self._prj_cfg['ProjectInfo']['didDefFile']

    def get_prj_config(self):
        """Get the project configuration name from the config file.

        Returns:
            Project config name

        """
        return self._prj_cfg['ProjectInfo']["projConfig"]

    def get_a2l_name(self):
        """Get the name of the a2l-file, which the build system shall generate."""
        return self._prj_cfg['ProjectInfo']['a2LFileName']

    def get_ecu_info(self):
        """Return ecuSupplier and ecuType.

        Returns:
            (ecuSupplier, ecuType)
        """
        return (
            self._prj_cfg['ProjectInfo'].get('ecuSupplier', None),
            self._prj_cfg['ProjectInfo'].get('ecuType', '')
        )

    def get_xcp_enabled(self):
        """Return True/False whether XCP is enabled in the project or not.

        Returns:
            (bool): True/False whether XCP is enabled in the project or not

        """
        return self._prj_cfg['ProjectInfo'].get('enableXcp', True)

    def get_nvm_defs(self):
        """Return NVM-ram block definitions.

        The definitions contains the sizes of the six NVM areas
        which are defined in the build-system.

        Returns:
        NvmConfig dict from config file.

        """
        return self._prj_cfg['NvmConfig']

    def _get_inc_dirs(self, path):
        """Get the dirs with the models defined in the units config file.

        Model name somewhere in the path.
        """
        all_dirs = glob.glob(path)
        inc_units = self.get_included_units()
        psep = os.path.sep
        out = {}
        for dir_ in all_dirs:
            folders = dir_.split(psep)
            for inc_unit in inc_units:
                if inc_unit in folders:
                    out.update({inc_unit: dir_})
                    break
        return out

    def get_units_raster_cfg(self):
        """Get the units' scheduling raster config.

        I.e. which units are included, and in which
        rasters they are scheduled, and in which order.

        Returns:
            A dict in the following format.

        ::

            {
              "SampleTimes": {
                "NameOfRaster": scheduling time},
               "Rasters": {
                 "NameOfRaster": [
                   "NameOfFunction",
                   ...],
                 ...}
            }

        Example::

            {
              "SampleTimes": {
                "Vc10ms": 0.01,
                "Vc40ms": 0.04},
               "Rasters": {
                 "Vc10ms": [
                   "VcPpmImob",
                   "VcPpmPsm",
                   "VcPpmRc",
                   "VcPpmSt",
                   "VcPemAlc"],
                 "Vc40ms": [
                   "VcRegCh"]
            }

        """
        return self._unit_cfg

    def get_unit_cfg_dirs(self):
        """Get config dirs which matches the project config parameter prjUnitCfgDir.

        Furthermore, they should be included in the unit definition for this project

        Returns:
            A list with absolute paths to all unit config dirs
            included in the project

        """
        path = os.path.join(self.get_root_dir(),
                            os.path.normpath(self._prj_cfg['ProjectInfo']
                                             ['prjUnitCfgDir']))
        return self._get_inc_dirs(path)

    def get_translation_files_dirs(self):
        """Get translation files directories, specified as a path regex in project
        config by key prjTranslationDir. If key is not present, will fall back to
        prjUnitCfgDir.

        Returns:
            A dictionary with absolute paths to all translation file dirs included
            in the project
        """

        if "prjTranslationDir" not in self._prj_cfg['ProjectInfo']:
            return self.get_unit_cfg_dirs()

        normpath_dir = os.path.normpath(self._prj_cfg['ProjectInfo']['prjTranslationDir'])
        path = os.path.join(self.get_root_dir(), normpath_dir)

        all_dirs = glob.glob(path)
        translation_dirs = {}
        for directory in all_dirs:
            file = pathlib.Path(directory).stem
            translation_dirs[file] = directory
        return translation_dirs

    def get_common_src_dir(self):
        """Get source dir which matches the project config parameter commonSrcDir.

        Returns:
            Absolute path to common source dir

        """
        return os.path.join(self.get_root_dir(),
                            os.path.normpath(self._prj_cfg['ProjectInfo']['commonSrcDir']))

    def get_unit_src_dirs(self):
        """Get source dirs which matches the project config parameter prjUnitCfgDir.

        Furthermore, they should be included in the unit definition for this project

        Returns:
            A list with absolute paths to all source dirs included in the
            project

        """
        path = os.path.join(self.get_root_dir(),
                            os.path.normpath(self._prj_cfg['ProjectInfo']
                                             ['prjUnitSrcDir']))
        return self._get_inc_dirs(path)

    def get_unit_mdl_dirs(self):
        """Get source dirs which matches the project config parameter prjUnitCfgDir.

        Furthermore, they should be included in the unit definition for this project

        Returns:
            A list with absolute paths to all model dirs included in the
            project

        """
        path = os.path.join(self.get_root_dir(),
                            os.path.normpath(self._prj_cfg['ProjectInfo']
                                             ['prjUnitMdlDir']))
        return self._get_inc_dirs(path)

    def get_use_global_const(self):
        """Get the name of the global constant module."""
        return self._prj_cfg['ProjectInfo']['useGlobalConst']

    def get_use_volatile_globals(self):
        """Get if global variables should be defined as volatile or not."""
        if 'useVolatileGlobals' in self._prj_cfg['ProjectInfo']:
            return self._prj_cfg['ProjectInfo']['useVolatileGlobals']
        return False

    def get_use_custom_dummy_spm(self):
        """Get path to file defining missing internal variables, if any.

        This file will be used instead of generating VcDummy_spm.c,
        to make it easier to maintain missing internal signals.

        Returns:
            customDummySpm (os.path): An absolute path to the custom dummy spm file, if existent.
        """
        if 'customDummySpm' in self._prj_cfg['ProjectInfo']:
            return os.path.join(
                self.get_root_dir(),
                os.path.normpath(self._prj_cfg['ProjectInfo']['customDummySpm'])
            )
        return None

    def get_use_custom_sources(self):
        """Get path to files with custom handwritten sourcecode, if any.

        Returns:
            customSources (os.path): A list of absolute paths to custom sources, if existent.
        """
        if 'customSources' in self._prj_cfg['ProjectInfo']:
            normalized_paths = (os.path.normpath(p) for p in self._prj_cfg['ProjectInfo']['customSources'])
            return [os.path.join(self.get_root_dir(), p) for p in normalized_paths]
        return None

    def get_if_cfg_dir(self):
        """Return the directory containing the interface configuration files.

        Returns:
            An absolute path to the interface configuration files

        """
        return os.path.join(self._prj_root_dir,
                            self._prj_cfg['ProjectInfo']['interfaceCfgDir'])

    def get_enum_def_dir(self):
        """Get path to dir containing simulink enumeration definitions, if any.

        Returns:
            enumDefDir (os.path): An absolute path to the simulink enumerations, if existent.
        """
        if 'enumDefDir' in self._prj_cfg['ProjectInfo']:
            return os.path.join(
                self.get_root_dir(),
                os.path.normpath(self._prj_cfg['ProjectInfo']['enumDefDir'])
            )
        return None


if __name__ == '__main__':
    # Function for testing the module
    BPC = BuildProjConfig('../../ProjectCfg.json')



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Module for handling ZoneController calibration."""

import re
from pathlib import Path
from powertrain_build.problem_logger import ProblemLogger


class ZoneControllerCalibration(ProblemLogger):
    """Class for handling ZoneController calibration."""

    calibration_function_init_template = '{swc_name}_ZcCalibrationInit'
    calibration_function_step_template = '{swc_name}_ZcCalibrationStep'
    trigger_read_rte_cdata_signal = {
        'name_template': 'c{swc_name}_TriggerReadRteCData',
        'data_type': 'UInt8'
    }

    def __init__(self, build_cfg, calib_data):
        """Init.

        Args:
            build_cfg (BuildProjConfig): Object with build configuration settings.
            calib_data (dict): Dictionary containing calibration data for a ZoneController project.
        """
        self.swc_name = build_cfg.get_composition_config("softwareComponentName")
        self.asil_level = re.sub("ASIL(?=[^_])", "ASIL_", build_cfg.get_composition_config("asil"))
        self.src_code_dst_dir = build_cfg.get_src_code_dst_dir()
        self.calibration_variables = calib_data['class_info']
        cal_interface_filename = f"calibration_interface_{build_cfg.get_scheduler_prefix()}".rstrip("_")
        self.calibration_interface_header = f'{cal_interface_filename}.h'
        self.calibration_interface_source = f'{cal_interface_filename}.c'
        self.trigger_read_rte_cdata_signal_name = self.trigger_read_rte_cdata_signal['name_template'].format(
            swc_name=self.swc_name
        )

    def _get_header_guard(self):
        header_guard_tmp = Path(self.calibration_interface_header).stem
        return header_guard_tmp.upper() + '_H'

    def _get_calibration_variables_write_string_list(self, indent=4):
        write_string_list = []
        for signal_name, signal_data in self.calibration_variables.items():
            rte_call = f'Rte_CData_{self.swc_name}_{signal_name}()'
            if isinstance(signal_data["width"], list):
                # MAPs get typedef:ed to structs and need special data type mapping
                if signal_name.startswith("m") and signal_name[-2:] not in ["_r", "_c"]:
                    write_string_list.append(
                        f'{"":{indent}}const {signal_data["autosar_type"]} *{signal_name}_ptr = {rte_call};\n'
                    )
                    write_string_list.append(
                        f'{"":{indent}}memcpy('
                        f'{signal_name}, '
                        f'{signal_name}_ptr->{signal_data["autosar_type"]}, '
                        f'sizeof({signal_name})'
                        ');\n'
                    )
                else:
                    write_string_list.append(
                        f'{"":{indent}}memcpy({signal_name}, {rte_call}, sizeof({signal_name}));\n'
                    )
            else:
                write_string_list.append(f'{"":{indent}}{signal_name} = {rte_call};\n')
        return write_string_list

    def _get_source_file_init_content(self):
        start_include = "CVC_CODE_START.h" if self.asil_level == "QM" else f"CVC_CODE_{self.asil_level}_START.h"
        stop_include = "CVC_CODE_END.h" if self.asil_level == "QM" else f"CVC_CODE_{self.asil_level}_END.h"
        header = [f'#include "{start_include}"\n']
        footer = [f'#include "{stop_include}"\n', '\n']

        body = [
            f'void {self.calibration_function_init_template.format(swc_name=self.swc_name)}(void)\n',
            '{\n',
        ]
        body.extend(self._get_calibration_variables_write_string_list())
        body.append('}\n')

        return header + body + footer

    def _get_source_file_step_content(self):
        start_include = "CVC_CODE_START.h" if self.asil_level == "QM" else f"CVC_CODE_{self.asil_level}_START.h"
        stop_include = "CVC_CODE_END.h" if self.asil_level == "QM" else f"CVC_CODE_{self.asil_level}_END.h"
        header = [f'#include "{start_include}"\n']
        footer = [f'#include "{stop_include}"\n']

        trigger_read_calibration_function = f'Rte_CData_{self.swc_name}_{self.trigger_read_rte_cdata_signal_name}()'
        body = [
            f'void {self.calibration_function_step_template.format(swc_name=self.swc_name)}(void)\n',
            '{\n',
            f'    if ({self.trigger_read_rte_cdata_signal_name} != {trigger_read_calibration_function})\n'
        ]
        body.append('    {\n')
        body.extend(self._get_calibration_variables_write_string_list(indent=8))
        body.append('    }\n')
        body.append('}\n')

        return header + body + footer

    def get_header_file_content(self):
        """Get content for the calibration header file.

        Returns:
            (list(str)): List of lines to write to calibration header file.
        """
        lines_to_write = []
        start_include = "PREDECL_CAL_START.h" if self.asil_level == "QM" else f"PREDECL_CAL_{self.asil_level}_START.h"
        stop_include = "PREDECL_CAL_END.h" if self.asil_level == "QM" else f"PREDECL_CAL_{self.asil_level}_END.h"
        header = [
            f'#ifndef {self._get_header_guard()}\n',
            f'#define {self._get_header_guard()}\n',
            '#define CVC_CAL\n',
            '#include <string.h>\n',
            '#include "tl_basetypes.h"\n',
            f'#include "Rte_{self.swc_name}.h"\n',
            '\n'
        ]
        footer = [
            '\n',
            f'#endif /* {self._get_header_guard()} */\n'
        ]

        lines_to_write.append(f'#include "{start_include}"\n')
        for signal_name, signal_data in self.calibration_variables.items():
            if isinstance(signal_data["width"], list):
                rows, cols = signal_data["width"]
                if rows > 1:
                    declaration = f'extern CVC_CAL {signal_data["type"]} {signal_name}[{rows}][{cols}];\n'
                else:
                    declaration = f'extern CVC_CAL {signal_data["type"]} {signal_name}[{cols}];\n'
            else:
                declaration = f'extern CVC_CAL {signal_data["type"]} {signal_name};\n'
            lines_to_write.append(declaration)
        lines_to_write.append(f'#include "{stop_include}"\n')

        lines_to_write.append('\n')
        for signal_name, signal_data in self.calibration_variables.items():
            if isinstance(signal_data["width"], list):
                # MAPs get typedef:ed to structs and need special data type mapping
                if signal_name.startswith("m") and signal_name[-2:] not in ["_r", "_c"]:
                    return_type = f'const {signal_data["autosar_type"]}*'
                else:
                    return_type = f'const {signal_data["type"]}*'
            else:
                return_type = f'{signal_data["type"]}'
            lines_to_write.append(f'extern {return_type} Rte_CData_{self.swc_name}_{signal_name}(void);\n')

        lines_to_write.extend([
            '\n',
            f'void {self.calibration_function_init_template.format(swc_name=self.swc_name)}(void);\n',
            f'void {self.calibration_function_step_template.format(swc_name=self.swc_name)}(void);\n'
        ])

        return header + lines_to_write + footer

    def get_source_file_content(self):
        """Get content for the calibration source file.

        Returns:
            (list(str)): List of lines to write to calibration source file.
        """
        start_include = "CVC_CAL_START.h" if self.asil_level == "QM" else f"CVC_CAL_{self.asil_level}_START.h"
        stop_include = "CVC_CAL_END.h" if self.asil_level == "QM" else f"CVC_CAL_{self.asil_level}_END.h"
        header = [
            f'#include "{self.calibration_interface_header}"\n',
            '\n',
            f'#include "{start_include}"\n',
            f'CVC_CAL {self.trigger_read_rte_cdata_signal["data_type"]} '
            f'{self.trigger_read_rte_cdata_signal_name} = 0;\n',
            f'#include "{stop_include}"\n',
            '\n',
        ]

        body = self._get_source_file_init_content()
        body.extend(self._get_source_file_step_content())

        return header + body

    def generate_calibration_interface_files(self):
        """Generate calibration interface files."""
        header_file_content = self.get_header_file_content()
        calibration_interface_header_path = Path(self.src_code_dst_dir, self.calibration_interface_header)
        with calibration_interface_header_path.open(mode='w', encoding='utf-8') as file_handler:
            file_handler.writelines(header_file_content)

        source_file_content = self.get_source_file_content()
        calibration_interface_source_path = Path(self.src_code_dst_dir, self.calibration_interface_source)
        with calibration_interface_source_path.open(mode='w', encoding='utf-8') as file_handler:
            file_handler.writelines(source_file_content)
